/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.probes.Probe;

/**
 * <p>
 * An abstract probe to manage the lifecycle.
 * </p>
 * 
 * @see {@link Probe}
 * @author Christophe DENEUX (PetalsLink)
 * 
 */
public abstract class AbstractProbe implements Probe {

    /**
     * <p>
     * Is the probe initialized ?
     * </p>
     */
    protected volatile boolean isInitialized = false;

    /**
     * <p>
     * Is the probe started ?
     * </p>
     */
    protected volatile boolean isStarted = false;

    /**
     * Lock to manage the lifecycle changes.
     */
    private Object lifecycleLock = new Object();

    protected AbstractProbe() {
        // NOP
    }

    @Override
    public final void init() throws ProbeInitializedException, ProbeStartedException,
            ProbeInitializationException {
        synchronized (this.lifecycleLock) {
            if (this.isStarted) {
                throw new ProbeStartedException();
            }
            if (this.isInitialized) {
                throw new ProbeInitializedException();
            }
            this.doInit();
            this.isInitialized = true;
        }
    }

    /**
     * <p>
     * Method to override to add processing on the initialization phase of the
     * probe.
     * </p>
     * <b>MUST</b> be thread safe.
     * 
     * @throws ProbeInitializationException
     *             An error occurs during the probe initialization.
     */
    public abstract void doInit() throws ProbeInitializationException;

    @Override
    public final void start() throws ProbeNotInitializedException, ProbeStartedException,
            ProbeStartupException {
        synchronized (this.lifecycleLock) {
            if (!this.isInitialized) {
                throw new ProbeNotInitializedException();
            }
            if (this.isStarted) {
                throw new ProbeStartedException();
            }

            this.doStart();
            this.isStarted = true;
        }
    }

    /**
     * <p>
     * Method to override to add processing on the start phase of the probe.
     * </p>
     * <b>MUST</b> be thread safe.
     * 
     * @throws ProbeStartupException
     *             An error occurs during the probe start.
     */
    public abstract void doStart() throws ProbeStartupException;

    @Override
    public final void stop() throws ProbeNotStartedException, ProbeStopException,
            ProbeNotInitializedException {
        synchronized (this.lifecycleLock) {
            if (!this.isInitialized) {
                throw new ProbeNotInitializedException();
            }
            if (!this.isStarted) {
                throw new ProbeNotStartedException();
            }

            try {
                this.doStop();
            } finally {
                this.isStarted = false;
            }
        }
    }

    /**
     * <p>
     * Method to override to add processing on the stop phase of the probe.
     * </p>
     * <b>MUST</b> be thread safe.
     * 
     * @throws ProbeStopException
     *             An error occurs during the probe stop.
     */
    public abstract void doStop() throws ProbeStopException;

    @Override
    public final void shutdown() throws ProbeShutdownException, ProbeStartedException,
            ProbeNotInitializedException {
        synchronized (this.lifecycleLock) {
            if (this.isStarted) {
                throw new ProbeStartedException();
            }
            if (!this.isInitialized) {
                throw new ProbeNotInitializedException();
            }

            try {
                this.doShutdown();
            } finally {
                this.isInitialized = false;
            }
        }
    }

    /**
     * <p>
     * Method to override to add processing on the shutdown phase of the probe.
     * </p>
     * <b>MUST</b> be thread safe.
     * 
     * @throws ProbeShutdownException
     *             An error occurs during the probe shutdown.
     */
    public abstract void doShutdown() throws ProbeShutdownException;
}
