/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import java.util.concurrent.atomic.AtomicLong;

import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.probes.CounterProbe;

/**
 * <p>
 * A simple counter probe to count the number of something since its last
 * start.</ul>
 * </p>
 * <p>
 * <u><b>Lifecycle</b>:</u> The counter value is incremented only if the probe
 * is started. Counter value can be acceded even if the probe is initialized or
 * stopped. Counter value is reset when the probe starts.
 * </p>
 * 
 * @author Christophe DENEUX (PetalsLink)
 * 
 */
public class CounterProbeImpl extends AbstractProbe implements CounterProbe {

    /**
     * Value of the counter
     */
    private final AtomicLong value = new AtomicLong(0);

    public CounterProbeImpl() {
        // NOP
    }

    /**
     * Increments the value of the counter.
     * 
     * @throws ProbeNotStartedException
     *             The probe is not started.
     */
    public void inc() throws ProbeNotStartedException {
        if (this.isStarted) {
            // Increment the counter
            this.value.incrementAndGet();
        } else {
            throw new ProbeNotStartedException();
        }
    }

    /**
     * Accessor to the counter value.
     * 
     * @return The counter value since the last probe start.
     * @throws ProbeNotInitializedException
     *             The probe is not initialized.
     */
    public long getValue() throws ProbeNotInitializedException {
        if (this.isInitialized) {
            return this.value.longValue();
        } else {
            throw new ProbeNotInitializedException();
        }
    }

    @Override
    public final void doInit() throws ProbeInitializationException {
        // NOP
    }

    @Override
    public final void doStart() throws ProbeStartupException {
        // Reset the counter value
        this.value.set(0);
    }

    @Override
    public final void doStop() throws ProbeStopException {
        // NOP
    }

    @Override
    public final void doShutdown() throws ProbeShutdownException {
        // NOP
    }
}
