/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.atomic.AtomicLong;

import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeKeyMissingException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.key.ProbeKey;
import org.ow2.petals.probes.api.probes.KeyedCounterProbe;
import org.ow2.petals.probes.api.sensor.detector.KeyedCounterDefectDetector;

/**
 * <p>
 * A counter probe to count the number of something by key (K) since its last
 * start.</ul>
 * </p>
 * <p>
 * <u><b>Lifecycle</b>:</u> The counter values are incremented only if the probe
 * is started. Counter values can be acceded even if the probe is initialized or
 * stopped. Counter values are reset when the probe starts or shutdowns.
 * </p>
 * 
 * @param <K>
 *            Type of the key identifying each counter, implementing
 *            {@link ProbeKey}.
 * 
 * @author Christophe DENEUX (PetalsLink)
 * 
 */
public class KeyedCounterProbeImpl<K extends ProbeKey> extends AbstractKeyedProbe<K> implements
        KeyedCounterProbe<K> {

    private final static long INITIAL_VALUE = 1L;

	/**
	 * Value of the counters
	 */
    private final ConcurrentMap<K, AtomicLong> values = new ConcurrentHashMap<K, AtomicLong>();

    private final KeyedCounterDefectDetector defectDetector;

    /**
     * Create a {@link KeyedCounterProbeImpl} without defect detector.
     */
    public KeyedCounterProbeImpl() {
        this(null);
    }

    /**
     * Create a {@link KeyedCounterProbeImpl} using a defect detector
     * 
     * @param defectDetector
     *            The defect detector, can be null.
     */
    public KeyedCounterProbeImpl(final KeyedCounterDefectDetector defectDetector) {
        super();
        this.defectDetector = defectDetector;
	}

	/**
	 * Increments the value of the counter.
	 * 
	 * @throws ProbeNotStartedException
	 *             The probe is not started.
	 */
	public void inc(final K key) throws ProbeNotStartedException {
		if (this.isStarted) {
			// Increment the counter
            final AtomicLong initialValueIfNeeded = new AtomicLong(INITIAL_VALUE);
			final AtomicLong value = this.values.putIfAbsent(key,
					initialValueIfNeeded);
			if (value != null) {
                final long primValue = value.incrementAndGet();
                if (this.defectDetector != null) {
                    this.defectDetector.detectOnIncrement(key, primValue);
                }
            } else {
                if (this.defectDetector != null) {
                    this.defectDetector.detectOnIncrement(key, INITIAL_VALUE);
                }
			}
		} else {
			throw new ProbeNotStartedException();
		}
	}

	/**
	 * Accessor to the counter values.
	 * 
	 * @return A copy of the counter values since the last probe start.
	 * @throws ProbeNotInitializedException
	 *             The probe is not initialized.
	 */
	public Map<K, Long> getValues() throws ProbeNotInitializedException {
		if (this.isInitialized) {
			final Map<K, Long> results = new HashMap<K, Long>(
					this.values.size());

            // 'getValues' MUST be synchronized with 'move'. A monitoring client
			// should not get values between decrement and increment.
			synchronized (this.values) {
				final Iterator<Map.Entry<K, AtomicLong>> itValues = this.values
						.entrySet().iterator();
				while (itValues.hasNext()) {
					final Map.Entry<K, AtomicLong> entry = itValues.next();
					results.put(entry.getKey(), entry.getValue().longValue());
				}
			}
			return results;
		} else {
			throw new ProbeNotInitializedException();
		}
	}
	
    public Map<String[], Long> getConvertedValues() throws ProbeNotInitializedException {

		return this.convertValues(this.getValues());
	}
	
	private Map<String[], Long> convertValues(final Map<K, Long> map) {
		final Map<String[], Long> results = new HashMap<String[], Long>(map.size());

		final Iterator<Map.Entry<K, Long>> itValues = map.entrySet().iterator();
		while (itValues.hasNext()) {
			final Map.Entry<K, Long> entry = itValues.next();
			results.put(entry.getKey().toStringArray(), entry.getValue());
		}
		return results;
	}

	/**
	 * Decrements the counter associated to the key <code>originalKey</code> and
	 * increments the counter associated to the key <code>newKey</code>.
	 * 
	 * @param originalKey
	 *            The key associated to the counter to decrement
	 * @param newKey
	 *            The key associated to the counter to increment
	 * @throws ProbeNotStartedException
	 *             The probe is not started.
	 * @throws ProbeKeyMissingException
	 *             The <code>originalKey</code>, or <code>newKey</code> does not exist in the probe.
	 */
	public void move(final K originalKey, final K newKey)
			throws ProbeKeyMissingException, ProbeNotStartedException {
		if (this.isStarted) {

			// 'move' MUST be synchronized with 'getValues' to be ATOMIC in
			// the point of view of the monitoring client. A monitoring client
			// should not get values between decrement and increment.
			synchronized (this.values) {
				final AtomicLong originalValue = this.values.get(originalKey);
				if (originalValue != null) {
					originalValue.decrementAndGet();
					
                    final AtomicLong initialValueIfNeeded = new AtomicLong(INITIAL_VALUE);
					final AtomicLong newValue = this.values.putIfAbsent(newKey, initialValueIfNeeded);
					if (newValue != null) {
						newValue.incrementAndGet();
					}
					else {
                        // NOP: The new key was not already in the list and it
                        // is initialized to INITIAL_VALUE
					}
				} else {
					throw new ProbeKeyMissingException(originalKey.toStringArray());
				}
			}
		} else {
			throw new ProbeNotStartedException();
		}
	}

	@Override
	public final void doInit() throws ProbeInitializationException {
		// NOP
	}

	@Override
	public final void doStart() throws ProbeStartupException {
		// Reset the counter values
		this.values.clear();
	}

	@Override
	public final void doStop() throws ProbeStopException {
		// NOP
	}

	@Override
	public final void doShutdown() throws ProbeShutdownException {
		// Reset the counter values
		this.values.clear();
	}
}
