/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.key.ProbeKey;
import org.ow2.petals.probes.api.probes.KeyedGaugeProbe;
import org.ow2.petals.probes.api.sensor.KeyedGaugeSensor;

/**
 * <p>
 * A gauge probe to probe instant value, by key (K), on demand, and storing the
 * max reached value (that can be greater than the max of instant values) since
 * its last start.</ul>
 * </p>
 * <p>
 * <u><b>Lifecycle</b>:</u> The instant values can be acceded or picked only if
 * the probe is started. The max values can be acceded even if the probe is
 * initialized or stopped. The max values are reseted to the sensor initial
 * value when the probe starts.
 * </p>
 * 
 * @param <K>
 *            Type of the key identifying each value, implementing
 *            {@link ProbeKey}.
 * @param <T>
 *            The type of measured values (internal value type).
 * @param <S>
 *            The type of value to return to client probe (external value type).
 * 
 * @author Christophe DENEUX (PetalsLink)
 * 
 */
public class KeyedGaugeProbeImpl<K extends ProbeKey, T extends Comparable<T>, S> extends
        AbstractKeyedProbe<K> implements KeyedGaugeProbe<K, T, S> {

	/**
	 * Max value of the probe
	 */
    private ConcurrentMap<K, T> maxValues = new ConcurrentHashMap<K, T>();

	/**
	 * <p>
	 * Synchronization object to access read/write the {@link #maxValue}.
	 * </p>
	 * <p>
	 * A dedicated object is needed because instance of {@link #maxValue} will
	 * be re-assigned when a new max value is detected. So the lock will not
	 * work correctly.
	 * </p>
	 */
	private final Object lockMaxValue = new Object();

	/**
	 * The sensor to access instant value
	 */
	protected final KeyedGaugeSensor<K, T, S> sensor;

    public KeyedGaugeProbeImpl(final KeyedGaugeSensor<K, T, S> sensor) {
		this.sensor = sensor;
	}

	/**
	 * <p>
	 * Realizes a measure and compare the value to the max one to update it.
	 * </p>
	 * <p>
	 * This method <b>MUST</b> be invoked by the application internals when the
	 * resource associated to the probe changes.
	 * </p>
	 * <p>
	 * This method is thread-safe
	 * </p>
	 * 
	 * @param key
	 *            The key for which the measure must be done.
	 * @throws ProbeNotStartedException
	 *             The probe is not started.
	 */
	public void pick(final K key) throws ProbeNotStartedException {
		this.localInstantValue(key);
	}

	/**
	 * <p>
	 * Realizes a measure and compare the value to the max one to update it.
	 * </p>
	 * <p>
	 * This method is thread-safe
	 * </p>
	 * 
	 * @param key
	 *            The key for which the measure must be done.
	 * @throws ProbeNotStartedException
	 *             The probe is not started.
	 */
	private T localInstantValue(final K key) throws ProbeNotStartedException {
		if (this.isStarted) {
			final T instantValue = this.sensor.getInstantValue(key);
			synchronized (this.lockMaxValue) {
				final T maxVal = this.maxValues.get(key);
				if (maxVal != null) {
					if (instantValue.compareTo(maxVal) > 0) {
						this.maxValues.put(key, instantValue);
					}
				} else {
					// It's a new key
					this.maxValues.put(key, instantValue);
				}
				return instantValue;
			}
		} else {
			throw new ProbeNotStartedException();
		}
	}

	/**
	 * <p>
	 * Accessor to the sensor instant values of all keys.
	 * </p>
	 * <p>
	 * This method <b>MUST</b> be invoked by monitoring clients to get the probe
	 * instant values.
	 * </p>
	 * <p>
	 * This method is thread-safe
	 * </p>
	 * 
	 * @return A copy of the sensor instant values associated to all keys intercepted
	 *         since the last probe start.
	 * @throws ProbeNotStartedException
	 *             The probe is not started.
	 */
	public Map<K, T> getInstantValues() throws ProbeNotStartedException {

		if (this.isStarted) {
			final Map<K, T> results = new HashMap<K, T>(
					this.maxValues.size());
			final Iterator<K> itValues = this.maxValues.keySet().iterator();
			while (itValues.hasNext()) {
				final K key = itValues.next();
				results.put(key, this.localInstantValue(key));
			}
			return results;

		} else {
			throw new ProbeNotStartedException();
		}
	}
	
	public Map<String[], S> getConvertedInstantValues() throws ProbeNotStartedException {

		return this.convertValues(this.getInstantValues());
	}
	
	private Map<String[], S> convertValues(final Map<K, T> map) {
		final Map<String[], S> results = new HashMap<String[], S>(map.size());

		final Iterator<Map.Entry<K, T>> itValues = map.entrySet().iterator();
		while (itValues.hasNext()) {
			final Map.Entry<K, T> entry = itValues.next();
			results.put(entry.getKey().toStringArray(), this.sensor.toExternalValue(entry.getValue()));
		}
		return results;
	}

	/**
	 * <p>
	 * Accessor to the max value.
	 * </p>
	 * <p>
	 * This method is thread-safe
	 * </p>
	 * 
	 * @return The probe max value.
	 * @throws ProbeNotInitializedException
	 *             The probe is not initialized.
	 */
	public Map<K, T> getMaxValues() throws ProbeNotInitializedException {
		if (this.isInitialized) {
			final Map<K, T> results = new HashMap<K, T>(this.maxValues.size());
			final Iterator<Map.Entry<K, T>> itValues = this.maxValues
					.entrySet().iterator();
			while (itValues.hasNext()) {
				final Map.Entry<K, T> entry = itValues.next();
				results.put(entry.getKey(), entry.getValue());
			}
			return results;

		} else {
			throw new ProbeNotInitializedException();
		}
	}
	
	public Map<String[], S> getConvertedMaxValues() throws ProbeNotInitializedException {

		return this.convertValues(this.getMaxValues());
	}

	@Override
	public final void doInit() throws ProbeInitializationException {
		// NOP
	}

	@Override
	public final void doStart() throws ProbeStartupException {
		// Reset the sensor
		this.sensor.reset();
		
		// Reset the max value
		synchronized (this.lockMaxValue) {
			this.maxValues.clear();
		}
	}

	@Override
	public final void doStop() throws ProbeStopException {
		// NOP
	}

	@Override
	public final void doShutdown() throws ProbeShutdownException {
		// Free resources
		this.sensor.reset();
		synchronized (this.lockMaxValue) {
			this.maxValues.clear();
		}		
	}
}
