/**
 * Copyright (c) 2013-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl.macro;

import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.probes.GaugeProbe;
import org.ow2.petals.probes.api.probes.macro.ObjectPoolProbe;
import org.ow2.petals.probes.api.sensor.detector.DefectCreator;
import org.ow2.petals.probes.impl.GaugeProbeImpl;
import org.ow2.petals.probes.impl.macro.detector.ObjectPoolBorrowedObjectsDefectDetector;
import org.ow2.petals.probes.impl.macro.sensor.ObjectPoolBorrowedObjectsGaugeSensor;
import org.ow2.petals.probes.impl.macro.sensor.ObjectPoolIdleObjectsGaugeSensor;

/**
 * <p>
 * A macro containing all probes needed to monitor a {@link GenericObjectPool}
 * </p>
 * <p>
 * <u><b>Lifecycle</b>:</u> The instant value can be acceded or picked only if
 * the probe is started. The max value can be acceded even if the probe is
 * initialized or stopped. The max value is reseted to the sensor initial value
 * when the probe starts.
 * </p>
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ObjectPoolProbeImpl implements ObjectPoolProbe {

    /**
     * The probe counting borrowed objects in the pool.
     */
    private GaugeProbe<Long, Long> probeObjectPoolBorrowedObjects;

    /**
     * The probe counting idle objects of the object pool.
     */
    private GaugeProbe<Long, Long> probeObjectPoolIdleObjects;

    /**
     * Sensor attached to the borrowed objects of the object pool.
     */
    private final ObjectPoolBorrowedObjectsGaugeSensor objectPoolBorrowedObjectsGaugeSensor;

    /**
     * Defect detector attached to the borrowed objects of the object pool.
     */
    private final ObjectPoolBorrowedObjectsDefectDetector objectPoolBorrowedObjectsDefectDetector;

    /**
     * Sensor attached to the idle objects of the object pool.
     */
    private final ObjectPoolIdleObjectsGaugeSensor objectPoolIdleObjectsGaugeSensor;

    /**
     * The probed generic object pool
     */
    private GenericObjectPool<?> objectPool = null;

    /**
     * @param objectPoolBorrowedObjectsDefectCreator
     *            Defect creator used to create defects about exhaustion of the
     *            object pool
     */
    public ObjectPoolProbeImpl(final DefectCreator objectPoolBorrowedObjectsDefectCreator) {

        // Create the probe about allocated threads
        this.objectPoolBorrowedObjectsGaugeSensor = new ObjectPoolBorrowedObjectsGaugeSensor();
        this.objectPoolBorrowedObjectsDefectDetector = new ObjectPoolBorrowedObjectsDefectDetector(
                objectPoolBorrowedObjectsDefectCreator);
        this.probeObjectPoolBorrowedObjects = new GaugeProbeImpl<Long, Long>(
                this.objectPoolBorrowedObjectsGaugeSensor,
                this.objectPoolBorrowedObjectsDefectDetector);

        // Create the probe about idle threads
        this.objectPoolIdleObjectsGaugeSensor = new ObjectPoolIdleObjectsGaugeSensor();
        this.probeObjectPoolIdleObjects = new GaugeProbeImpl<Long, Long>(
                this.objectPoolIdleObjectsGaugeSensor, null);

    }

    @Override
    public void setObjectPool(final GenericObjectPool<?> objectPool) {
        this.objectPool = objectPool;
        this.objectPoolBorrowedObjectsDefectDetector.setObjectPool(this.objectPool);
        this.objectPoolBorrowedObjectsGaugeSensor.setObjectPool(this.objectPool);
        this.objectPoolIdleObjectsGaugeSensor.setObjectPool(this.objectPool);
    }

    @Override
    public long getObjectPoolMaxSize() {
        if (this.objectPool == null) {
            return 0L;
        } else {
            return this.objectPool.getMaxActive();
        }
    }

    @Override
    public long getObjectPoolMinIdleSize() {
        if (this.objectPool == null) {
            return 0L;
        } else {
            return this.objectPool.getMinIdle();
        }
    }

    @Override
    public long getObjectPoolBorrowedObjectsMax() throws ProbeNotInitializedException {
        return this.probeObjectPoolBorrowedObjects.getMaxValue();
    }

    @Override
    public long getObjectPoolBorrowedObjectsCurrent() throws ProbeNotStartedException {
        return this.probeObjectPoolBorrowedObjects.getInstantValue();
    }

    @Override
    public long getObjectPoolIdleObjectsMax() throws ProbeNotInitializedException {
        return this.probeObjectPoolIdleObjects.getMaxValue();
    }

    @Override
    public long getObjectPoolIdleObjectsCurrent() throws ProbeNotStartedException {
        return this.probeObjectPoolIdleObjects.getInstantValue();
    }

    @Override
    public void pick() throws ProbeNotStartedException {
        this.probeObjectPoolBorrowedObjects.pick();
        this.probeObjectPoolIdleObjects.pick();
    }

    @Override
    public void init() throws ProbeInitializedException, ProbeStartedException,
            ProbeInitializationException {
        this.probeObjectPoolBorrowedObjects.init();
        this.probeObjectPoolIdleObjects.init();
    }

    @Override
    public void start() throws ProbeNotInitializedException, ProbeStartedException,
            ProbeStartupException {
        this.probeObjectPoolBorrowedObjects.start();
        this.probeObjectPoolIdleObjects.start();
    }

    @Override
    public void stop() throws ProbeNotInitializedException, ProbeNotStartedException,
            ProbeStopException {
        this.probeObjectPoolBorrowedObjects.stop();
        this.probeObjectPoolIdleObjects.stop();
    }

    @Override
    public void shutdown() throws ProbeShutdownException, ProbeStartedException,
            ProbeNotInitializedException {
        this.probeObjectPoolBorrowedObjects.shutdown();
        this.probeObjectPoolIdleObjects.shutdown();
    }

}
