/**
 * Copyright (c) 2014-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import static org.junit.Assert.assertEquals;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Semaphore;

import org.junit.Test;
import org.ow2.petals.probes.api.exceptions.ProbeException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.probes.CounterGaugeProbe;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * <p>
 * Unit tests of {@link CounterGaugeProbe} to verify:
 * <ul>
 * <li>the counting,</li>
 * <li>the thread-safety</li>
 * <li>specific methods against the lifecycle.</li>
 * </ul>
 * </p>
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class CounterGaugeProbeImplTestCase {

    private final static short THREAD_POOL_SIZE = 200;

    private final static long THREADSAFETY_ITERATION_NB = 10000;

    /**
     * <p>
     * Tests the method 'inc'/'dec' of a started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the counter is incremented/decremented,</li>
     * <li>min and max values are correct.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testIncAndDec_001() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();

        counterGaugeProbe.init();
        counterGaugeProbe.start();
        counterGaugeProbe.inc();
        assertEquals("The counter should be incremented.", 1, counterGaugeProbe.getCurrentValue());
        counterGaugeProbe.inc();
        assertEquals("The counter should be incremented.", 2, counterGaugeProbe.getCurrentValue());
        counterGaugeProbe.dec();
        assertEquals("The counter should be decremented.", 1, counterGaugeProbe.getCurrentValue());
        counterGaugeProbe.dec();
        assertEquals("The counter should be decremented.", 0, counterGaugeProbe.getCurrentValue());
        assertEquals("Invalid max value", 2, counterGaugeProbe.getMaxValue());
        assertEquals("Invalid min value", 0, counterGaugeProbe.getMinValue());
    }

    /**
     * <p>
     * Tests the method 'inc' of a re-started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the counter is incremented during the first startup,</li>
     * <li>the counter is reset when restarting the probe,</li>
     * <li>the counter is incremented during the 2nd startup.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testInc_002() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();

        counterGaugeProbe.init();
        counterGaugeProbe.start();
        counterGaugeProbe.inc();
        assertEquals("The counter should be incremented.", 1, counterGaugeProbe.getCurrentValue());
        counterGaugeProbe.stop();
        counterGaugeProbe.start();
        assertEquals("The counter should be reseted.", 0, counterGaugeProbe.getCurrentValue());
        counterGaugeProbe.inc();
        assertEquals("The counter should be incremented.", 1, counterGaugeProbe.getCurrentValue());
    }

    /**
     * <p>
     * Tests the method 'dec' of a re-started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the counter is decremented during the first startup,</li>
     * <li>the counter is reset when restarting the probe,</li>
     * <li>the counter is decremented during the 2nd startup.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testDec_002() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();

        counterGaugeProbe.init();
        counterGaugeProbe.start();
        counterGaugeProbe.dec();
        assertEquals("The counter should be decremented.", -1, counterGaugeProbe.getCurrentValue());
        counterGaugeProbe.stop();
        counterGaugeProbe.start();
        assertEquals("The counter should be reseted.", 0, counterGaugeProbe.getCurrentValue());
        counterGaugeProbe.dec();
        assertEquals("The counter should be decremented.", -1, counterGaugeProbe.getCurrentValue());
    }

    /**
     * <p>
     * Tests the method {@link CounterGaugeProbeImpl#getCurrentValue()} of a just initialized probe.
     * </p>
     * <p>
     * Expected results: an error occurs because the probe is not started.
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testGetCurrentValue_001() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();

        counterGaugeProbe.init();
        counterGaugeProbe.getCurrentValue();
    }

    /**
     * <p>
     * Tests the method {@link CounterGaugeProbeImpl#getMinValue()} of a just initialized probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the counter has its initial value.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetMinValue_001() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();

        counterGaugeProbe.init();
        assertEquals("The counter has not its initial value.", 0, counterGaugeProbe.getMinValue());
    }

    /**
     * <p>
     * Tests the method {@link CounterGaugeProbeImpl#getMaxValue()} of a just initialized probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the counter has its initial value.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetMaxValue_001() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();

        counterGaugeProbe.init();
        assertEquals("The counter has not its initial value.", 0, counterGaugeProbe.getMaxValue());
    }

    /**
     * <p>
     * Tests the method {@link CounterGaugeProbeImpl#getCurrentValue()} of a started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>just after the startup, the counter has its initial value,</li>
     * <li>the counter has the right value after incrementing,</li>
     * <li>the counter has the right value after decrementing,</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetCurrentValue_002() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();

        counterGaugeProbe.init();
        counterGaugeProbe.start();
        assertEquals("The counter has not its initial value.", 0, counterGaugeProbe.getCurrentValue());
        counterGaugeProbe.inc();
        assertEquals("The counter has not the right value after incrementing.", 1, counterGaugeProbe.getCurrentValue());
        counterGaugeProbe.dec();
        assertEquals("The counter has not the right value after decrementing.", 0, counterGaugeProbe.getCurrentValue());
        assertEquals("The counter has not the right max value after inc/dec.", 1, counterGaugeProbe.getMaxValue());
        assertEquals("The counter has not the right min value after inc/dec.", 0, counterGaugeProbe.getMinValue());
    }

    /**
     * <p>
     * Tests the method {@link CounterGaugeProbeImpl#getCurrentValue()} of a stopped probe.
     * </p>
     * <p>
     * Expected results: An error occurs because the probe is stopped. </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testGetCurrentValue_003() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();


        counterGaugeProbe.init();
        counterGaugeProbe.start();
        counterGaugeProbe.inc();
        counterGaugeProbe.dec();
        counterGaugeProbe.stop();
        assertEquals("The counter has not the right max value after inc/dec.", 1, counterGaugeProbe.getMaxValue());
        assertEquals("The counter has not the right min value after inc/dec.", 0, counterGaugeProbe.getMinValue());
        counterGaugeProbe.getCurrentValue();
    }

    /*----------------------------*/
    /* Lifecycle error test cases */
    /*----------------------------*/

    /**
     * <p>
     * Tests the method 'inc' of a just allocated probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>the counter is not incremented.</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testInc_101() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();

        try {
            counterGaugeProbe.inc();
        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
            assertEquals("The counter should not be incremented.", 0, ((Long) ReflectionHelper.getPrivateFieldValue(
                    CounterGaugeProbeImpl.class, counterGaugeProbe, "instantValue")).longValue());
            throw e;
        }
    }

    /**
     * <p>
     * Tests the method 'dec' of a just allocated probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>the counter is not incremented.</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testDec_101() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();

        try {
            counterGaugeProbe.dec();
        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
            assertEquals("The counter should not be decremented.", 0, ((Long) ReflectionHelper.getPrivateFieldValue(
                    CounterGaugeProbeImpl.class, counterGaugeProbe, "instantValue")).longValue());
            throw e;
        }
    }

    /**
     * <p>
     * Tests the method 'inc' of a stop probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>the counter is not incremented.</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testInc_102() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();

        try {
            counterGaugeProbe.init();
            counterGaugeProbe.start();
            counterGaugeProbe.inc();
            counterGaugeProbe.stop();

            counterGaugeProbe.inc();

        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
            assertEquals("The counter should not be incremented.", 1, ((Long) ReflectionHelper.getPrivateFieldValue(
                    CounterGaugeProbeImpl.class, counterGaugeProbe, "instantValue")).longValue());
            throw e;
        }
    }

    /**
     * <p>
     * Tests the method 'dec' of a stop probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>the counter is not incremented.</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testDec_102() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();

        try {
            counterGaugeProbe.init();
            counterGaugeProbe.start();
            counterGaugeProbe.dec();
            counterGaugeProbe.stop();

            counterGaugeProbe.dec();

        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
            assertEquals("The counter should not be decremented.", -1, ((Long) ReflectionHelper.getPrivateFieldValue(
                    CounterGaugeProbeImpl.class, counterGaugeProbe, "instantValue")).longValue());
            throw e;
        }
    }

    /**
     * <p>
     * Tests the method 'inc' of a shutdown probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>the counter is not incremented.</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testInc_103() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();

        try {
            counterGaugeProbe.init();
            counterGaugeProbe.start();
            counterGaugeProbe.inc();
            counterGaugeProbe.stop();
            counterGaugeProbe.shutdown();

            counterGaugeProbe.inc();

        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
            assertEquals("The counter should not be incremented.", 1, ((Long) ReflectionHelper.getPrivateFieldValue(
                    CounterGaugeProbeImpl.class, counterGaugeProbe, "instantValue")).longValue());
            throw e;
        }
    }

    /**
     * <p>
     * Tests the method 'dec' of a shutdown probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>the counter is not incremented.</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testDec_103() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();

        try {
            counterGaugeProbe.init();
            counterGaugeProbe.start();
            counterGaugeProbe.dec();
            counterGaugeProbe.stop();
            counterGaugeProbe.shutdown();

            counterGaugeProbe.dec();

        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
            assertEquals("The counter should not be decremented.", -1, ((Long) ReflectionHelper.getPrivateFieldValue(
                    CounterGaugeProbeImpl.class, counterGaugeProbe, "instantValue")).longValue());
            throw e;
        }
    }

    /**
     * <p>
     * Tests the method 'getCurrentValue' of a shutdown probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotInitializedException} is thrown</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testGetValue_101() throws ProbeException {
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();

        counterGaugeProbe.init();
        counterGaugeProbe.start();
        counterGaugeProbe.stop();
        counterGaugeProbe.shutdown();

        counterGaugeProbe.getCurrentValue();
    }

    /*--------------------------*/
    /* Thread-safety test cases */
    /*--------------------------*/

    /**
     * <p>
     * Tests the thread-safety of the methods 'inc' and 'getCurrentValue'. No change is made on the probe lifecycle.
     * </p>
     * 
     * @throws InterruptedException
     */
    @Test
    public void testThreadSafety_001() throws ProbeException, InterruptedException {

        // Testcase failures
        final List<Throwable> testcaseFailures = Collections
                .synchronizedList(new LinkedList<Throwable>());

        // Synchronization objects
        final Object starterMonitor = new Object();
        final Semaphore readyMonitor = new Semaphore(THREAD_POOL_SIZE);

        // Initialize and start the probe
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();
        counterGaugeProbe.init();
        counterGaugeProbe.start();

        // Initialize and start the thread pool size to THREAD_POOL_SIZE
        final List<Thread> threads = new LinkedList<Thread>();
        for (int i = 0; i < THREAD_POOL_SIZE; i++) {
            readyMonitor.acquire();

            final Thread thread = new Thread(new Runnable() {

                @Override
                public void run() {

                    try {
                        synchronized (starterMonitor) {
                            readyMonitor.release();
                            starterMonitor.wait();
                        }

                        for (int j = 0; j < THREADSAFETY_ITERATION_NB; j++) {
                            counterGaugeProbe.inc();
                            counterGaugeProbe.getCurrentValue();
                        }
                    } catch (final Throwable e) {
                        testcaseFailures.add(e);
                    }

                }
            });
            threads.add(thread);
            thread.start();
        }

        // Wait that threads end to start
        readyMonitor.acquire(THREAD_POOL_SIZE);

        synchronized (starterMonitor) {
            starterMonitor.notifyAll();
        }

        // Wait the end of processing
        for (Thread thread : threads) {
            thread.join();
        }

        assertEquals("Error occurs: " + testcaseFailures.toString(), 0, testcaseFailures.size());
        assertEquals("The expected counter value is not the expected one",
                THREAD_POOL_SIZE * THREADSAFETY_ITERATION_NB, counterGaugeProbe.getMaxValue());

    }

    /**
     * <p>
     * Tests the thread-safety of the methods 'dec' and 'getCurrentValue'. No change is made on the probe lifecycle.
     * </p>
     * 
     * @throws InterruptedException
     */
    @Test
    public void testThreadSafety_002() throws ProbeException, InterruptedException {

        // Testcase failures
        final List<Throwable> testcaseFailures = Collections.synchronizedList(new LinkedList<Throwable>());

        // Synchronization objects
        final Object starterMonitor = new Object();
        final Semaphore readyMonitor = new Semaphore(THREAD_POOL_SIZE);

        // Initialize and start the probe
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();
        counterGaugeProbe.init();
        counterGaugeProbe.start();

        // Initialize and start the thread pool size to THREAD_POOL_SIZE
        final List<Thread> threads = new LinkedList<Thread>();
        for (int i = 0; i < THREAD_POOL_SIZE; i++) {
            readyMonitor.acquire();

            final Thread thread = new Thread(new Runnable() {

                @Override
                public void run() {

                    try {
                        synchronized (starterMonitor) {
                            readyMonitor.release();
                            starterMonitor.wait();
                        }

                        for (int j = 0; j < THREADSAFETY_ITERATION_NB; j++) {
                            counterGaugeProbe.dec();
                            counterGaugeProbe.getCurrentValue();
                        }
                    } catch (final Throwable e) {
                        testcaseFailures.add(e);
                    }

                }
            });
            threads.add(thread);
            thread.start();
        }

        // Wait that threads end to start
        readyMonitor.acquire(THREAD_POOL_SIZE);

        synchronized (starterMonitor) {
            starterMonitor.notifyAll();
        }

        // Wait the end of processing
        for (Thread thread : threads) {
            thread.join();
        }

        assertEquals("Error occurs: " + testcaseFailures.toString(), 0, testcaseFailures.size());
        assertEquals("The expected counter value is not the expected one",
                0 - (THREAD_POOL_SIZE * THREADSAFETY_ITERATION_NB), counterGaugeProbe.getMinValue());

    }

    /**
     * <p>
     * Tests the thread-safety of the methods 'inc'/'dec' and 'getCurrentValue'. No change is made on the probe
     * lifecycle.
     * </p>
     * 
     * @throws InterruptedException
     */
    @Test
    public void testThreadSafety_003() throws ProbeException, InterruptedException {

        // Testcase failures
        final List<Throwable> testcaseFailures = Collections.synchronizedList(new LinkedList<Throwable>());

        // Synchronization objects
        final Object starterMonitor = new Object();
        final Semaphore readyMonitor = new Semaphore(THREAD_POOL_SIZE);

        // Initialize and start the probe
        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();
        counterGaugeProbe.init();
        counterGaugeProbe.start();

        // Initialize and start the thread pool size to THREAD_POOL_SIZE
        final List<Thread> threads = new LinkedList<Thread>();
        for (int i = 0; i < THREAD_POOL_SIZE; i++) {
            readyMonitor.acquire();

            final Thread thread = new Thread(new Runnable() {

                @Override
                public void run() {

                    try {
                        synchronized (starterMonitor) {
                            readyMonitor.release();
                            starterMonitor.wait();
                        }

                        for (int j = 0; j < THREADSAFETY_ITERATION_NB; j++) {
                            counterGaugeProbe.inc();
                            counterGaugeProbe.getCurrentValue();
                            counterGaugeProbe.dec();
                        }
                    } catch (final Throwable e) {
                        testcaseFailures.add(e);
                    }

                }
            });
            threads.add(thread);
            thread.start();
        }

        // Wait that threads end to start
        readyMonitor.acquire(THREAD_POOL_SIZE);

        synchronized (starterMonitor) {
            starterMonitor.notifyAll();
        }

        // Wait the end of processing
        for (Thread thread : threads) {
            thread.join();
        }

        assertEquals("Error occurs: " + testcaseFailures.toString(), 0,
                testcaseFailures.size());
        assertEquals("The expected counter value is not the expected one", 0, counterGaugeProbe.getCurrentValue());

    }

    /**
     * Check that no error occurs on start/stop cycles (see PETALSCDK-116)
     */
    @Test
    public void startStopCycles() throws ProbeInitializedException, ProbeStartedException,
            ProbeInitializationException, ProbeNotInitializedException, ProbeNotStartedException, ProbeStopException,
            ProbeStartupException, ProbeShutdownException {

        final CounterGaugeProbe counterGaugeProbe = new CounterGaugeProbeImpl();
        counterGaugeProbe.init();

        counterGaugeProbe.start();
        counterGaugeProbe.stop();

        counterGaugeProbe.start();
        counterGaugeProbe.stop();

        counterGaugeProbe.shutdown();
    }
    
    // TODO: Add a test about thread-safety of inc and lifecycle changes
}
