/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicLong;

import junit.framework.TestCase;

import org.junit.Test;
import org.ow2.petals.probes.api.exceptions.ProbeException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.probes.CounterProbe;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * <p>
 * Unit tests of {@link CounterProbe} to verify:
 * <ul>
 * <li>the counting,</li>
 * <li>the thread-safety</li>
 * <li>specific methods against the lifecycle.</li>
 * </ul>
 * </p>
 * 
 * @author Christophe DENEUX (PetalsLink)
 * 
 */
public class CounterProbeImplTestCase {

    private final static short THREAD_POOL_SIZE = 100;

    private final static long THREADSAFETY_ITERATION_NB = 10000;

    /**
     * <p>
     * Tests the method 'inc' of a started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the counter is incremented.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testInc_001() {
        final CounterProbe counterProbe = new CounterProbeImpl();

        try {
            counterProbe.init();
            counterProbe.start();
            counterProbe.inc();
            TestCase.assertEquals("The counter should be incremented.", 1, counterProbe.getValue());

        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'inc' of a re-started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the counter is incremented during the first startup,</li>
     * <li>the counter is reset when restarting the probe,</li>
     * <li>the counter is incremented during the 2nd startup.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testInc_002() {
        final CounterProbe counterProbe = new CounterProbeImpl();

        try {
            counterProbe.init();
            counterProbe.start();
            counterProbe.inc();
            TestCase.assertEquals("The counter should be incremented.", 1, counterProbe.getValue());
            counterProbe.stop();
            counterProbe.start();
            TestCase.assertEquals("The counter should be reseted.", 0, counterProbe.getValue());
            counterProbe.inc();
            TestCase.assertEquals("The counter should be incremented.", 1, counterProbe.getValue());

        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'getValue' of a just initialized probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>the counter has its initial value.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetValue_001() {
        final CounterProbe counterProbe = new CounterProbeImpl();

        try {
            counterProbe.init();
            TestCase.assertEquals("The counter has not its initial value.", 0,
                    counterProbe.getValue());

        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'getValue' of a started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>just after the startup, the counter has its initial value,</li>
     * <li>the counter has the right value after incrementing.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetValue_002() {
        final CounterProbe counterProbe = new CounterProbeImpl();

        try {
            counterProbe.init();
            counterProbe.start();
            TestCase.assertEquals("The counter has not its initial value.", 0,
                    counterProbe.getValue());
            counterProbe.inc();
            TestCase.assertEquals("The counter has not the right value after incrementing.", 1,
                    counterProbe.getValue());

        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'getValue' of a stopped probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>just after the startup, the counter has its initial value,</li>
     * <li>the counter has the right value after incrementing.</li>
     * <li>after to be stop, the counter has as the same value as before the
     * stop.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetValue_003() {
        final CounterProbe counterProbe = new CounterProbeImpl();

        try {
            counterProbe.init();
            counterProbe.start();
            TestCase.assertEquals("The counter has not its initial value.", 0,
                    counterProbe.getValue());
            counterProbe.inc();
            TestCase.assertEquals("The counter has not the right value after incrementing.", 1,
                    counterProbe.getValue());
            counterProbe.stop();
            TestCase.assertEquals("The counter has not the same value than before the stop.", 1,
                    counterProbe.getValue());

        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /*----------------------------*/
    /* Lifecycle error test cases */
    /*----------------------------*/

    /**
     * <p>
     * Tests the method 'inc' of a just allocated probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>the counter is not incremented.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testInc_101() {
        final CounterProbe counterProbe = new CounterProbeImpl();

        try {
            counterProbe.inc();
            TestCase.fail("The incrementation of a just allocated probe should fail with ProbeNotStartedException");
        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
            TestCase.assertEquals("The counter should not be incremented.", 0,
                    ((AtomicLong)ReflectionHelper.getPrivateFieldValue(CounterProbeImpl.class, counterProbe, "value")).longValue());
        }
    }

    /**
     * <p>
     * Tests the method 'inc' of a stop probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>the counter is not incremented.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testInc_102() {
        final CounterProbe counterProbe = new CounterProbeImpl();

        try {
            counterProbe.init();
            counterProbe.start();
            counterProbe.inc();
            counterProbe.stop();

            counterProbe.inc();
            TestCase.fail("The incrementation of a stopped probe should fail with ProbeNotStartedException");
        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
            TestCase.assertEquals("The counter should not be incremented.", 1,
                    ((AtomicLong)ReflectionHelper.getPrivateFieldValue(CounterProbeImpl.class, counterProbe, "value")).longValue());
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'inc' of a shutdown probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>the counter is not incremented.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testInc_103() {
        final CounterProbe counterProbe = new CounterProbeImpl();

        try {
            counterProbe.init();
            counterProbe.start();
            counterProbe.inc();
            counterProbe.stop();
            counterProbe.shutdown();

            counterProbe.inc();
            TestCase.fail("The incrementation of a shutdown probe should fail with ProbeNotStartedException");
        } catch (final ProbeNotStartedException e) {
            // NOP, it's the expected exception
            TestCase.assertEquals("The counter should not be incremented.", 1,
                    ((AtomicLong)ReflectionHelper.getPrivateFieldValue(CounterProbeImpl.class, counterProbe, "value")).longValue());
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /**
     * <p>
     * Tests the method 'getValue' of a shutdown probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotInitializedException} is thrown</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetValue_101() {
        final CounterProbe counterProbe = new CounterProbeImpl();

        try {
            counterProbe.init();
            counterProbe.start();
            counterProbe.stop();
            counterProbe.shutdown();

            counterProbe.getValue();
            TestCase.fail("The method 'getValue' should fail with ProbeNotInitializedException");
        } catch (final ProbeNotInitializedException e) {
            // NOP, it's the expected exception
        } catch (final ProbeException e) {
            TestCase.fail(e.getMessage());
        }
    }

    /*--------------------------*/
    /* Thread-safety test cases */
    /*--------------------------*/

    /**
     * <p>
     * Tests the thread-safety of the methods 'inc' and 'getValue'. No change is
     * made on the probe lifecycle.
     * </p>
     * 
     * @throws InterruptedException
     */
    @Test
    public void testThreadSafety_001() throws ProbeException, InterruptedException {

        // Testcase failures
        final List<Throwable> testcaseFailures = Collections
                .synchronizedList(new LinkedList<Throwable>());

        // Synchronization objects
        final Object starterMonitor = new Object();
        final Semaphore readyMonitor = new Semaphore(THREAD_POOL_SIZE);

        // Initialize and start the probe
        final CounterProbe counterProbe = new CounterProbeImpl();
        counterProbe.init();
        counterProbe.start();

        // Initialize and start the thread pool size to THREAD_POOL_SIZE
        final List<Thread> threads = new LinkedList<Thread>();
        for (int i = 0; i < THREAD_POOL_SIZE; i++) {
            readyMonitor.acquire();

            final Thread thread = new Thread(new Runnable() {

                @Override
                public void run() {

                    try {
                        synchronized (starterMonitor) {
                            readyMonitor.release();
                            starterMonitor.wait();
                        }

                        for (int j = 0; j < THREADSAFETY_ITERATION_NB; j++) {
                            counterProbe.inc();
                            counterProbe.getValue();
                        }
                    } catch (final Throwable e) {
                        testcaseFailures.add(e);
                    }

                }
            });
            threads.add(thread);
            thread.start();
        }

        // Wait that threads end to start
        readyMonitor.acquire(THREAD_POOL_SIZE);

        synchronized (starterMonitor) {
            starterMonitor.notifyAll();
        }

        // Wait the end of processing
        for (Thread thread : threads) {
            thread.join();
        }

        TestCase.assertEquals("Error occurs: " + testcaseFailures.toString(), 0,
                testcaseFailures.size());
        TestCase.assertEquals("The expected counter value is not the expected one",
                THREAD_POOL_SIZE * THREADSAFETY_ITERATION_NB, counterProbe.getValue());

    }

    /**
     * Check that no error occurs on start/stop cycles (see PETALSCDK-116)
     */
    @Test
    public void startStopCycles() throws ProbeInitializedException, ProbeStartedException,
            ProbeInitializationException, ProbeNotInitializedException, ProbeNotStartedException, ProbeStopException,
            ProbeStartupException, ProbeShutdownException {

        final CounterProbe counterProbe = new CounterProbeImpl();
        counterProbe.init();

        counterProbe.start();
        counterProbe.stop();

        counterProbe.start();
        counterProbe.stop();

        counterProbe.shutdown();
    }
    
    // TODO: Add a test about thread-safety of inc and lifecycle changes
}
