/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicLong;

import junit.framework.TestCase;

import org.junit.Test;
import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.ProbeException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeKeyMissingException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.probes.KeyedCounterProbe;
import org.ow2.petals.probes.key.ProbeTestExecutionStatusKey;
import org.ow2.petals.probes.key.ProbeTestKey;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * <p>
 * Unit tests of {@link KeyedCounterProbe} to verify:
 * <ul>
 * <li>the counting by key,</li>
 * <li>the thread-safety</li>
 * <li>specific methods against the lifecycle.</li>
 * </ul>
 * </p>
 * 
 * @author Christophe DENEUX (PetalsLink)
 * 
 */
public class KeyedCounterProbeImplTestCase {

	private final static short THREAD_POOL_SIZE = 25;

	private final static long THREADSAFETY_ITERATION_NB = 10000;

	private final static short KEYS_NUMBER = 4;

	/**
	 * <p>
	 * Tests the method 'inc' of a started probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>no exception is thrown,</li>
	 * <li>the counters are correctly incremented.</li>
	 * </ul>
	 * </p>
	 */
    @Test
	public void testInc_001() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
		final ProbeTestKey key1 = new ProbeTestKey("key1");
		final ProbeTestKey key2 = new ProbeTestKey("key2");

		try {
			counterProbe.init();
			counterProbe.start();
			counterProbe.inc(key1);
			counterProbe.inc(key1);
			counterProbe.inc(key2);

			TestCase.assertEquals(
					"The counter 'key1' has not the expected value.", 2,
					counterProbe.getValues().get(key1).longValue());
			TestCase.assertEquals(
					"The counter 'key2' has not the expected value.", 1,
					counterProbe.getValues().get(key2).longValue());

		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'inc' of a re-started probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>no exception is thrown,</li>
	 * <li>the counter are correctly incremented during the first startup,</li>
	 * <li>the counter are reset when restarting the probe,</li>
	 * <li>the counter are incremented during the 2nd startup.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testInc_002() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
		final ProbeTestKey key1 = new ProbeTestKey("key1");
		final ProbeTestKey key2 = new ProbeTestKey("key2");

		try {
			counterProbe.init();
			counterProbe.start();
			counterProbe.inc(key1);
			counterProbe.inc(key1);
			counterProbe.inc(key2);
			TestCase.assertEquals(
					"The counter 'key1' has not the expected value.", 2,
					counterProbe.getValues().get(key1).longValue());
			TestCase.assertEquals(
					"The counter 'key2' has not the expected value.", 1,
					counterProbe.getValues().get(key2).longValue());
			counterProbe.stop();
			counterProbe.start();
			TestCase.assertEquals(
					"At least one counter is available in the probe.", 0,
					counterProbe.getValues().size());
			counterProbe.inc(key1);
			counterProbe.inc(key2);
			counterProbe.inc(key2);
			TestCase.assertEquals(
					"The counter 'key1' has not the expected value.", 1,
					counterProbe.getValues().get(key1).longValue());
			TestCase.assertEquals(
					"The counter 'key2' has not the expected value.", 2,
					counterProbe.getValues().get(key2).longValue());

		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'getValues' of a just initialized probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>no exception is thrown,</li>
	 * <li>no counter is available in the probe.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testGetValues_001() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();

		try {
			counterProbe.init();
			TestCase.assertEquals(
					"At least one counter is available in the probe.", 0,
					counterProbe.getValues().size());

		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'getValues' of a started probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>no exception is thrown,</li>
	 * <li>just after the startup, no counter is available in the probe,</li>
	 * <li>the counters have the right value after incrementing several keys.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testGetValues_002() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
		final ProbeTestKey key1 = new ProbeTestKey("key1");
		final ProbeTestKey key2 = new ProbeTestKey("key2");

		try {
			counterProbe.init();
			counterProbe.start();
			TestCase.assertEquals(
					"At least one counter is available in the probe.", 0,
					counterProbe.getValues().size());
			counterProbe.inc(key1);
			counterProbe.inc(key1);
			counterProbe.inc(key2);

			TestCase.assertEquals(
					"The counter 'key1' has not the expected value.", 2,
					counterProbe.getValues().get(key1).longValue());
			TestCase.assertEquals(
					"The counter 'key2' has not the expected value.", 1,
					counterProbe.getValues().get(key2).longValue());

		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'getValues' of a stopped probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>no exception is thrown,</li>
	 * <li>just after the startup, no counter is available in the probe,</li>
	 * <li>the counters have the right value after incrementing.</li>
	 * <li>after to be stop, the counters have as the same value as before the
	 * stop.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testGetValues_003() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
		final ProbeTestKey key1 = new ProbeTestKey("key1");
		final ProbeTestKey key2 = new ProbeTestKey("key2");

		try {
			counterProbe.init();
			counterProbe.start();
			TestCase.assertEquals(
					"At least one counter is available in the probe.", 0,
					counterProbe.getValues().size());
			counterProbe.inc(key1);
			counterProbe.inc(key1);
			counterProbe.inc(key2);
			TestCase.assertEquals(
					"The counter 'key1' has not the expected value.", 2,
					counterProbe.getValues().get(key1).longValue());
			TestCase.assertEquals(
					"The counter 'key2' has not the expected value.", 1,
					counterProbe.getValues().get(key2).longValue());
			counterProbe.stop();
			TestCase.assertEquals(
					"The counter 'key1' has not the expected value after the probe stop.",
					2, counterProbe.getValues().get(key1).longValue());
			TestCase.assertEquals(
					"The counter 'key2' has not the expected value after the probe stop.",
					1, counterProbe.getValues().get(key2).longValue());

		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'move' of a started probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>no exception is thrown,</li>
	 * <li>the counters are correctly incremented and moved.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testMove_001() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
		final ProbeTestKey key1 = new ProbeTestKey("key1");
		final ProbeTestKey key2 = new ProbeTestKey("key2");

		try {
			counterProbe.init();
			counterProbe.start();
			counterProbe.inc(key1);
			counterProbe.inc(key2);
			TestCase.assertEquals(
					"The counter 'key1' has not the expected value.", 1,
					counterProbe.getValues().get(key1).longValue());
			TestCase.assertEquals(
					"The counter 'key2' has not the expected value.", 1,
					counterProbe.getValues().get(key2).longValue());

			counterProbe.move(key1, key2);
			TestCase.assertEquals(
					"The counter 'key1' has not the expected value.", 0,
					counterProbe.getValues().get(key1).longValue());
			TestCase.assertEquals(
					"The counter 'key2' has not the expected value.", 2,
					counterProbe.getValues().get(key2).longValue());

		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'move' of a started probe. A key is moved to a key that
	 * does not exist yet.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>no exception is thrown,</li>
	 * <li>the counters are correctly incremented and moved.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testMove_002() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
		final ProbeTestKey key1 = new ProbeTestKey("key1");

		try {
			counterProbe.init();
			counterProbe.start();
			counterProbe.inc(key1);
			TestCase.assertEquals(
					"The counter 'key1' has not the expected value.", 1,
					counterProbe.getValues().get(key1).longValue());

			final ProbeTestKey key2 = new ProbeTestKey("key2");
			counterProbe.move(key1, key2);
			TestCase.assertEquals(
					"The counter 'key1' has not the expected value.", 0,
					counterProbe.getValues().get(key1).longValue());
			TestCase.assertEquals(
					"The counter 'key2' has not the expected value.", 1,
					counterProbe.getValues().get(key2).longValue());

		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/*----------------------------*/
	/* Lifecycle error test cases */
	/*----------------------------*/

	/**
	 * <p>
	 * Tests the method 'inc' of a just allocated probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
	 * <li>no counter is available in the probe.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testInc_101() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
		final ProbeTestKey key1 = new ProbeTestKey("key1");

		try {
			counterProbe.inc(key1);
			TestCase.fail("The incrementation by key of a just allocated probe should fail with ProbeNotStartedException");
		} catch (final ProbeNotStartedException e) {
			// NOP, it's the expected exception
			TestCase.assertEquals(
					"At least one counter is available in the probe.", 0,
					((Map<?, AtomicLong>) ReflectionHelper.getPrivateFieldValue(KeyedCounterProbeImpl.class, counterProbe, "values")).size());
		}
	}

	/**
	 * <p>
	 * Tests the method 'inc' of a stop probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
	 * <li>after to be stop, the counters have as the same value as before the
	 * stop.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testInc_102() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
		final ProbeTestKey key1 = new ProbeTestKey("key1");

		try {
			counterProbe.init();
			counterProbe.start();
			counterProbe.inc(key1);
			counterProbe.stop();

			counterProbe.inc(key1);
			TestCase.fail("The incrementation of a stopped probe should fail with ProbeNotStartedException");
		} catch (final ProbeNotStartedException e) {
			// NOP, it's the expected exception
			TestCase.assertEquals("The counter should not be incremented.", 1,
                    ((Map<?, AtomicLong>) ReflectionHelper.getPrivateFieldValue(KeyedCounterProbeImpl.class, counterProbe, "values")).get(key1).longValue());
		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'inc' of a shutdown probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
	 * <li>no counter is available in the probe.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testInc_103() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
		final ProbeTestKey key1 = new ProbeTestKey("key1");

		try {
			counterProbe.init();
			counterProbe.start();
			counterProbe.inc(key1);
			counterProbe.stop();
			counterProbe.shutdown();

			counterProbe.inc(key1);
			TestCase.fail("The incrementation of a shutdown probe should fail with ProbeNotStartedException");
		} catch (final ProbeNotStartedException e) {
			// NOP, it's the expected exception
			TestCase.assertEquals(
					"At least one counter is available in the probe.", 0,
					((Map<?, AtomicLong>) ReflectionHelper.getPrivateFieldValue(KeyedCounterProbeImpl.class, counterProbe, "values")).size());
		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'getValues' of a shutdown probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotInitializedException} is thrown</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testGetValues_101() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();

		try {
			counterProbe.init();
			counterProbe.start();
			counterProbe.stop();
			counterProbe.shutdown();

			counterProbe.getValues();
			TestCase.fail("The method 'getValues' should fail with ProbeNotInitializedException");
		} catch (final ProbeNotInitializedException e) {
			// NOP, it's the expected exception
		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'move' of a just allocated probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
	 * <li>no counter is available in the probe.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testMove_101() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
		final ProbeTestKey key1 = new ProbeTestKey("key1");
		final ProbeTestKey key2 = new ProbeTestKey("key2");

		try {
			counterProbe.move(key1, key2);
			TestCase.fail("Moving one item from key1 to key2 of a just allocated probe should fail with ProbeNotStartedException");
		} catch (final ProbeNotStartedException e) {
			// NOP, it's the expected exception
			TestCase.assertEquals(
					"At least one counter is available in the probe.", 0,
					((Map<?, AtomicLong>) ReflectionHelper.getPrivateFieldValue(KeyedCounterProbeImpl.class, counterProbe, "values")).size());
		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'move' of a stopped probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
	 * <li>after to be stop, the counters have as the same value as before the
	 * stop.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testMove_102() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
		final ProbeTestKey key1 = new ProbeTestKey("key1");
		final ProbeTestKey key2 = new ProbeTestKey("key2");

		try {
			counterProbe.init();
			counterProbe.start();
			counterProbe.inc(key1);
			counterProbe.inc(key2);
			TestCase.assertEquals(
					"The counter 'key1' has not the expected value.", 1,
					counterProbe.getValues().get(key1).longValue());
			TestCase.assertEquals(
					"The counter 'key2' has not the expected value.", 1,
					counterProbe.getValues().get(key2).longValue());
			counterProbe.stop();
			counterProbe.move(key1, key2);
			TestCase.fail("Moving one item from key1 to key2 of a stopped probe should fail with ProbeNotStartedException");
		} catch (final ProbeNotStartedException e) {
			// NOP, it's the expected exception
			TestCase.assertEquals(
					"The counter 'key1' has not the expected value.", 1,
					((Map<?, AtomicLong>) ReflectionHelper.getPrivateFieldValue(KeyedCounterProbeImpl.class, counterProbe, "values")).get(key1).longValue());
			TestCase.assertEquals(
					"The counter 'key2' has not the expected value.", 1,
					((Map<?, AtomicLong>) ReflectionHelper.getPrivateFieldValue(KeyedCounterProbeImpl.class, counterProbe, "values")).get(key2).longValue());
		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'move' of a shutdown probe.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
	 * <li>no counter is available in the probe.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testMove_103() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
		final ProbeTestKey key1 = new ProbeTestKey("key1");
		final ProbeTestKey key2 = new ProbeTestKey("key2");

		try {
			counterProbe.init();
			counterProbe.start();
			counterProbe.inc(key1);
			counterProbe.inc(key2);
			TestCase.assertEquals(
					"The counter 'key1' has not the expected value.", 1,
					counterProbe.getValues().get(key1).longValue());
			TestCase.assertEquals(
					"The counter 'key2' has not the expected value.", 1,
					counterProbe.getValues().get(key2).longValue());
			counterProbe.stop();
			counterProbe.shutdown();
			counterProbe.move(key1, key2);
			TestCase.fail("Moving one item from key1 to key2 of a shutdown probe should fail with ProbeNotStartedException");
		} catch (final ProbeNotStartedException e) {
			// NOP, it's the expected exception
			TestCase.assertEquals(
					"At least one counter is available in the probe.", 0,
					((Map<?, AtomicLong>) ReflectionHelper.getPrivateFieldValue(KeyedCounterProbeImpl.class, counterProbe, "values")).size());
		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/**
	 * <p>
	 * Tests the method 'move' of a started probe, using an unknown key as
	 * source key.
	 * </p>
	 * <p>
	 * Expected results:
	 * <ul>
	 * <li>the exception {@link ProbeKeyMissingException} is thrown,</li>
	 * <li>no change is done on counters.</li>
	 * </ul>
	 * </p>
	 */
    @Test
    public void testMove_104() {
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
		final ProbeTestKey key1 = new ProbeTestKey("key1");
		final ProbeTestKey key2 = new ProbeTestKey("key2");
		final ProbeTestKey unknownKey = new ProbeTestKey("unknownKey");

		try {
			counterProbe.init();
			counterProbe.start();
			counterProbe.inc(key1);
			counterProbe.inc(key2);
			TestCase.assertEquals(
					"The counter 'key1' has not the expected value.", 1,
					counterProbe.getValues().get(key1).longValue());
			TestCase.assertEquals(
					"The counter 'key2' has not the expected value.", 1,
					counterProbe.getValues().get(key2).longValue());

			counterProbe.move(unknownKey, key2);
			TestCase.fail("Moving one item from an unknown key should fail with ProbeKeyMissingException");

		} catch (final ProbeKeyMissingException e) {
			// NOP, it's the expected exception
			TestCase.assertEquals(
					"The counter 'key1' has not the expected value.", 1,
					((Map<?, AtomicLong>) ReflectionHelper.getPrivateFieldValue(KeyedCounterProbeImpl.class, counterProbe, "values")).get(key1).longValue());
			TestCase.assertEquals(
					"The counter 'key2' has not the expected value.", 1,
					((Map<?, AtomicLong>) ReflectionHelper.getPrivateFieldValue(KeyedCounterProbeImpl.class, counterProbe, "values")).get(key2).longValue());
		} catch (final ProbeException e) {
			TestCase.fail(e.getMessage());
		}
	}

	/*--------------------------*/
	/* Thread-safety test cases */
	/*--------------------------*/

	/**
	 * <p>
	 * Tests the thread-safety of the methods 'inc' and 'getValues'. No change
	 * is made on the probe lifecycle.
	 * </p>
	 * 
	 * @throws InterruptedException
	 */
    @Test
    public void testThreadSafety_001() throws ProbeException,
			InterruptedException {

		// Testcase failures
		final List<Throwable> testcaseFailures = Collections
				.synchronizedList(new LinkedList<Throwable>());

		// Synchronization objects
		final Object starterMonitor = new Object();
		final Semaphore readyMonitor = new Semaphore(THREAD_POOL_SIZE);

		// Initialize and start the probe
        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
		counterProbe.init();
		counterProbe.start();

		// Initialize and start the thread pool size to THREAD_POOL_SIZE
		final List<Thread> threads = new LinkedList<Thread>();
		for (int k = 0; k < KEYS_NUMBER; k++) {
			final ProbeTestKey key = new ProbeTestKey("key" + k);
			for (int i = 0; i < THREAD_POOL_SIZE; i++) {
				readyMonitor.acquire();

				final Thread thread = new Thread(new Runnable() {

					@Override
					public void run() {

						try {
							synchronized (starterMonitor) {
								readyMonitor.release();
								starterMonitor.wait();
							}

							for (int j = 0; j < THREADSAFETY_ITERATION_NB; j++) {
								counterProbe.inc(key);
								counterProbe.getValues();
							}
						} catch (final Throwable e) {
							testcaseFailures.add(e);
						}

					}
				});
				threads.add(thread);
				thread.start();
			}
		}

		// Wait that threads end to start
		readyMonitor.acquire(THREAD_POOL_SIZE);

		synchronized (starterMonitor) {
			starterMonitor.notifyAll();
		}

		// Wait the end of processing
		for (Thread thread : threads) {
			thread.join();
		}

		TestCase.assertEquals("Error occurs: " + testcaseFailures.toString(),
				0, testcaseFailures.size());

		for (int k = 0; k < KEYS_NUMBER; k++) {
			TestCase.assertEquals("The expected counter value of key 'key" + k
					+ "' is not the expected one", THREAD_POOL_SIZE
					* THREADSAFETY_ITERATION_NB,
					counterProbe.getValues().get(new ProbeTestKey("key" + k))
							.longValue());
		}
	}

	/**
	 * <p>
	 * Tests the thread-safety of the methods 'inc' and 'move'. No change is
	 * made on the probe lifecycle.
	 * </p>
	 * 
	 * @throws InterruptedException
	 */
    @Test
    public void testThreadSafety_002() throws ProbeException,
			InterruptedException {

		// Testcase failures
		final List<Throwable> testcaseFailures = Collections
				.synchronizedList(new LinkedList<Throwable>());

		// Synchronization objects
		final Object starterMonitor = new Object();
		final Semaphore readyMonitor = new Semaphore(THREAD_POOL_SIZE);

		// Initialize and start the probe
        final KeyedCounterProbe<ProbeTestExecutionStatusKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestExecutionStatusKey>();
		counterProbe.init();
		counterProbe.start();

		final BlockingQueue<ProbeTestExecutionStatusKey> incQueue = new LinkedBlockingQueue<ProbeTestExecutionStatusKey>();
		final BlockingQueue<ProbeTestExecutionStatusKey> moveQueue = new LinkedBlockingQueue<ProbeTestExecutionStatusKey>();

		// Initialize 'inc' command
		for (int j = 0; j < THREADSAFETY_ITERATION_NB * THREAD_POOL_SIZE; j++) {
			for (int k = 0; k < KEYS_NUMBER; k++) {
				final ProbeTestExecutionStatusKey key = new ProbeTestExecutionStatusKey(
						"key" + k, ExecutionStatus.PENDING);
				incQueue.add(key);
			}
		}

		// Initialize and start the thread pool size to THREAD_POOL_SIZE
		final List<Thread> threads = new LinkedList<Thread>();
		for (int i = 0; i < THREAD_POOL_SIZE; i++) {
			readyMonitor.acquire();

			// Threads that call 'inc' and initialize the 'move' command.
			final Thread incThread = new Thread(new Runnable() {

				@Override
				public void run() {

					try {
						synchronized (starterMonitor) {
							readyMonitor.release();
							starterMonitor.wait();
						}

						for (int j = 0; j < THREADSAFETY_ITERATION_NB
								* KEYS_NUMBER; j++) {
							final ProbeTestExecutionStatusKey key = incQueue
									.poll();
							counterProbe.inc(key);
							moveQueue.add(key);
						}

					} catch (final Throwable e) {
						testcaseFailures.add(e);
					}
				}
			});
			threads.add(incThread);
			incThread.start();

			// Threads that call 'move'.
			final Thread moveThread = new Thread(new Runnable() {

				@Override
				public void run() {

					try {
						synchronized (starterMonitor) {
							readyMonitor.release();
							starterMonitor.wait();
						}

						int j = 0;
						while (j < THREADSAFETY_ITERATION_NB * KEYS_NUMBER) {
							final ProbeTestExecutionStatusKey key = moveQueue.poll();
							if (key != null) {
								counterProbe.move(key, new ProbeTestExecutionStatusKey(key.getName(), ExecutionStatus.SUCCEEDED));
								j++;
							}
						}

					} catch (final Throwable e) {
						testcaseFailures.add(e);
					}
				}
			});
			threads.add(moveThread);
			moveThread.start();
		}

		// Wait that threads end to start
		readyMonitor.acquire(THREAD_POOL_SIZE * 2);

		synchronized (starterMonitor) {
			starterMonitor.notifyAll();
		}

		// Wait the end of processing
		for (Thread thread : threads) {
			thread.join();
		}

		if (testcaseFailures.size() > 0) {
			for (final Throwable t : testcaseFailures) {
				t.printStackTrace(System.err);
			}
		}
		TestCase.assertEquals("Error occurs: " + testcaseFailures.toString(),
				0, testcaseFailures.size());

		for (int k = 0; k < KEYS_NUMBER; k++) {
			final long counterValue = counterProbe.getValues().get(new ProbeTestExecutionStatusKey("key" + k, ExecutionStatus.SUCCEEDED)).longValue();
			System.out.println("Counter value for 'key" + k + ":" + ExecutionStatus.SUCCEEDED.toString() + "': " + counterValue);
			TestCase.assertEquals("The expected counter value of key 'key" + k
					+ "' is not the expected one", THREADSAFETY_ITERATION_NB * THREAD_POOL_SIZE,
					counterValue);
		}
	}

    /**
     * Check that no error occurs on start/stop cycles (see PETALSCDK-116)
     */
    @Test
    public void startStopCycles() throws ProbeInitializedException, ProbeStartedException,
            ProbeInitializationException, ProbeNotInitializedException, ProbeNotStartedException, ProbeStopException,
            ProbeStartupException, ProbeShutdownException {

        final KeyedCounterProbe<ProbeTestKey> counterProbe = new KeyedCounterProbeImpl<ProbeTestKey>();
        counterProbe.init();

        counterProbe.start();
        counterProbe.stop();

        counterProbe.start();
        counterProbe.stop();

        counterProbe.shutdown();
    }

	// TODO: Add tests about thread-safety of inc, move and getValues

	// TODO: Add tests about thread-safety of inc and lifecycle changes

	// TODO: Add tests about 'getConvertedInstantValues'
}
