/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.concurrent.Semaphore;

import org.junit.Test;
import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.ProbeException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.exceptions.StartDateItemUnknownException;
import org.ow2.petals.probes.api.key.ExecutionStatusProbeKey;
import org.ow2.petals.probes.api.key.ProbeKey;
import org.ow2.petals.probes.api.probes.AbsoluteResponseTimes;
import org.ow2.petals.probes.api.probes.KeyedResponseTimeProbe;
import org.ow2.petals.probes.api.probes.KeyedStartDateItem;
import org.ow2.petals.probes.api.probes.ResponseTimeRelativeValues;
import org.ow2.petals.probes.key.ExecutionStatusProbeKeyImpl;
import org.ow2.petals.probes.key.ProbeTestKey;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * <p>
 * Unit tests of {@link KeyedResponseTimeProbe} to verify:
 * <ul>
 * <li>the response time values computation by key,</li>
 * <li>the thread-safety</li>
 * <li>specific methods against the lifecycle.</li>
 * </ul>
 * </p>
 * 
 * @author Christophe DENEUX (PetalsLink)
 * 
 */
public class KeyedResponseTimeProbeImplTestCase {

    private final static short THREAD_POOL_SIZE = 25;

    private final static long THREADSAFETY_ITERATION_NB = 10000;

    private final static short KEYS_NUMBER = 4;

    private final static long SAMPLE_PERIOD = 5000;

    /**
     * Check that an error occurs when instantiating a
     * {@link KeyedResponseTimeProbeImpl} with a null timer.
     */
    @Test(expected = IllegalArgumentException.class)
    public void testConstructor_000() throws IllegalArgumentException {
        new KeyedResponseTimeProbeImpl<ProbeTestKey>(null,
                KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
    }

    /**
     * Check that an error occurs when instantiating a
     * {@link KeyedResponseTimeProbeImpl} with a sample period set to 0.
     */
    @Test(expected = IllegalArgumentException.class)
    public void testConstructor_001() throws IllegalArgumentException {
        final Timer timer = new Timer();
        try {
            new KeyedResponseTimeProbeImpl<ProbeTestKey>(timer, 0);
        } finally {
            timer.cancel();
        }
    }

    /**
     * Check that an error occurs when instantiating a
     * {@link KeyedResponseTimeProbeImpl} with a sample period set to a negative
     * value.
     */
    @Test(expected = IllegalArgumentException.class)
    public void testConstructor_002() throws IllegalArgumentException {
        final Timer timer = new Timer();
        try {
            new KeyedResponseTimeProbeImpl<ProbeTestKey>(timer, -250);
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'newExecution' of a started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>during the first sample:
     * <ul></li>each response time is correctly collected</li>
     * <li>no absolute response time values is available</li>
     * <li>no relative response time values is available</li>
     * </ul>
     * </li>
     * <li>after the switch of samples, relative and absolute response time
     * values are available.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testNewExecution_001() throws InterruptedException, ProbeException {
        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");
        final ProbeTestKey key2 = new ProbeTestKey("key2");

        try {
            responseTimeProbe.init();
            responseTimeProbe.start();

            responseTimeProbe.newExecution(key1);
            responseTimeProbe.newExecution(key1);
            responseTimeProbe.newExecution(key2);

            assertEquals(
                    "No absolute values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getAbsoluteResponseTimeValues().size());
            assertEquals(
                    "No relative values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getRelativeResponseTimeValues().size());

            // Wait the sample switches
            Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 1.1));

            assertEquals("Relative values should be available after sample switch.", 2,
                    responseTimeProbe.getRelativeResponseTimeValues().size());
            assertEquals("Absolute values should be available after sample switch.", 2,
                    responseTimeProbe.getAbsoluteResponseTimeValues().size());

        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'endsExecution' of a started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>during the first sample:
     * <ul></li>each response time is correctly collected</li>
     * <li>no absolute response time values is available</li>
     * <li>no relative response time values is available</li>
     * </ul>
     * </li>
     * <li>after the switch of samples, relative and absolute response time
     * values are available.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testEndsExecution_001() throws InterruptedException, ProbeException {
        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");
        final ProbeTestKey key2 = new ProbeTestKey("key2");

        try {
            responseTimeProbe.init();
            responseTimeProbe.start();

            final KeyedStartDateItem<ProbeTestKey> date1 = responseTimeProbe.newExecution(key1);
            final KeyedStartDateItem<ProbeTestKey> date2 = responseTimeProbe.newExecution(key1);
            final KeyedStartDateItem<ProbeTestKey> date3 = responseTimeProbe.newExecution(key2);

            // As it is possible to have endsExecution(...) invoked too fast
            // (<0,5ms), we force a sleeping wait
            Thread.sleep(500l);

            responseTimeProbe.endsExecution(date1, ExecutionStatus.SUCCEEDED);
            responseTimeProbe.endsExecution(date2, ExecutionStatus.SUCCEEDED);
            responseTimeProbe.endsExecution(date3, ExecutionStatus.SUCCEEDED);

            assertEquals(
                    "No absolute values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getAbsoluteResponseTimeValues().size());
            assertEquals(
                    "No relative values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getRelativeResponseTimeValues().size());

            // Wait more than one sample switch
            Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2.1));

            assertEquals(
                    "No relative values should be available because no execution is pending after sample switches.",
                    0, responseTimeProbe.getRelativeResponseTimeValues().size());
            assertEquals("Absolute values should be available after sample switch.", 2,
                    responseTimeProbe.getAbsoluteResponseTimeValues().size());
            final AbsoluteResponseTimes absRespTimes = responseTimeProbe
                    .getAbsoluteResponseTimeValues().get(
                            new ExecutionStatusProbeKeyImpl<ProbeKey>(key1,
                                    ExecutionStatus.SUCCEEDED));
            assertNotNull("No response time values for (key1, SUCCEEDED)", absRespTimes);
            assertTrue("The MIN of (key1, SUCCEEDED) is 0",
                    absRespTimes.getMinResponseTime() != 0L);
            assertTrue("The AVG of (key1, SUCCEEDED) is 0",
                    absRespTimes.getAvgResponseTime() != 0L);
            assertTrue("The MAX of (key1, SUCCEEDED) is 0",
                    absRespTimes.getMaxResponseTime() != 0L);

        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getAbsoluteResponseTimeValues' of an initialized probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>during the first sample: no absolute response time values is
     * available because no execution has been added,</li>
     * <li>after the switch of samples, no absolute response time values is
     * available because no execution has been added.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetAbsoluteResponseTimeValues_001() throws InterruptedException, ProbeException {
        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

        try {
            responseTimeProbe.init();

            assertEquals(
                    "No absolute values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getAbsoluteResponseTimeValues().size());

            // Wait the sample switch
            Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 1.1));

            assertEquals(
                    "No absolute values should be available because no execution has been added.",
                    0, responseTimeProbe.getAbsoluteResponseTimeValues().size());

        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getAbsoluteResponseTimeValues' of a started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>during the first sample:
     * <ul></li>each response time is correctly collected</li>
     * <li>no absolute response time values is available</li>
     * <li>no relative response time values is available</li>
     * </ul>
     * </li>
     * <li>after the switch of samples, expected absolute response time values
     * are available.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetAbsoluteResponseTimeValues_002() throws InterruptedException, ProbeException {
        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");
        final ProbeTestKey key2 = new ProbeTestKey("key2");

        try {
            responseTimeProbe.init();
            responseTimeProbe.start();

            final KeyedStartDateItem<ProbeTestKey> date1 = responseTimeProbe.newExecution(key1);
            responseTimeProbe.newExecution(key1);
            final KeyedStartDateItem<ProbeTestKey> date3 = responseTimeProbe.newExecution(key2);
            final KeyedStartDateItem<ProbeTestKey> date4 = responseTimeProbe.newExecution(key2);
            final KeyedStartDateItem<ProbeTestKey> date5 = responseTimeProbe.newExecution(key1);

            responseTimeProbe.endsExecution(date1, ExecutionStatus.SUCCEEDED);
            responseTimeProbe.endsExecution(date3, ExecutionStatus.ERROR);
            responseTimeProbe.endsExecution(date4, ExecutionStatus.FAULT);

            assertEquals(
                    "No absolute values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getAbsoluteResponseTimeValues().size());
            assertEquals(
                    "No relative values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getRelativeResponseTimeValues().size());

            // Wait sample switches
            Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2));
            responseTimeProbe.endsExecution(date5, ExecutionStatus.SUCCEEDED);
            Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2));

            final Map<ExecutionStatusProbeKey<ProbeTestKey>, AbsoluteResponseTimes> absoluteValues = responseTimeProbe
                    .getAbsoluteResponseTimeValues();
            assertEquals("Absolute values should be available after sample switch.", 4,
                    absoluteValues.size());
            assertNotNull("key1 with successfully execution is missing.",
                    absoluteValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key1, ExecutionStatus.SUCCEEDED)));
            assertTrue(
                    "Max response time of key1 with successfully execution is not the expected one.",
                    absoluteValues
                            .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key1,
                                    ExecutionStatus.SUCCEEDED)).getMaxResponseTime() >= KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2);

            assertTrue(
                    "Min response time of key1 with successfully execution is not the expected one.",
                    absoluteValues
                            .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key1,
                                    ExecutionStatus.SUCCEEDED)).getMinResponseTime() <= KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
            assertNotNull("key1 with pending execution is missing.",
                    absoluteValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key1, ExecutionStatus.PENDING)));
            assertNotNull("key2 with execution in error is missing.",
                    absoluteValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key2, ExecutionStatus.ERROR)));
            assertNotNull("key2 with faulty execution is missing.",
                    absoluteValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key2, ExecutionStatus.FAULT)));

        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getAbsoluteResponseTimeValues' of a stopped probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>during the first sample:
     * <ul></li>each response time is correctly collected</li>
     * <li>no absolute response time values is available</li>
     * <li>no relative response time values is available</li>
     * </ul>
     * </li>
     * <li>after the switch of samples, expected absolute response time values
     * are available.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetAbsoluteResponseTimeValues_003() throws InterruptedException, ProbeException {
        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");
        final ProbeTestKey key2 = new ProbeTestKey("key2");

        try {
            responseTimeProbe.init();
            responseTimeProbe.start();

            final KeyedStartDateItem<ProbeTestKey> date1 = responseTimeProbe.newExecution(key1);
            responseTimeProbe.newExecution(key1);
            final KeyedStartDateItem<ProbeTestKey> date3 = responseTimeProbe.newExecution(key2);
            final KeyedStartDateItem<ProbeTestKey> date4 = responseTimeProbe.newExecution(key2);
            final KeyedStartDateItem<ProbeTestKey> date5 = responseTimeProbe.newExecution(key1);

            responseTimeProbe.endsExecution(date1, ExecutionStatus.SUCCEEDED);
            responseTimeProbe.endsExecution(date3, ExecutionStatus.ERROR);
            responseTimeProbe.endsExecution(date4, ExecutionStatus.FAULT);

            assertEquals(
                    "No absolute values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getAbsoluteResponseTimeValues().size());
            assertEquals(
                    "No relative values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getRelativeResponseTimeValues().size());

            // Wait more than one sample switch
            Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2));
            responseTimeProbe.endsExecution(date5, ExecutionStatus.SUCCEEDED);
            responseTimeProbe.stop();
            Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2));

            final Map<ExecutionStatusProbeKey<ProbeTestKey>, AbsoluteResponseTimes> absoluteValues = responseTimeProbe
                    .getAbsoluteResponseTimeValues();
            assertEquals("Absolute values should be available after sample switch.", 4,
                    absoluteValues.size());
            assertNotNull("key1 with successfully execution is missing.",
                    absoluteValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key1, ExecutionStatus.SUCCEEDED)));
            assertTrue(
                    "Max response time of key1 with successfully execution is not the expected one.",
                    absoluteValues
                            .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key1,
                                    ExecutionStatus.SUCCEEDED)).getMaxResponseTime() >= KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2);

            assertTrue(
                    "Min response time of key1 with successfully execution is not the expected one.",
                    absoluteValues
                            .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key1,
                                    ExecutionStatus.SUCCEEDED)).getMinResponseTime() <= KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
            assertNotNull("key1 with pending execution is missing.",
                    absoluteValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key1, ExecutionStatus.PENDING)));
            assertNotNull("key2 with execution in error is missing.",
                    absoluteValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key2, ExecutionStatus.ERROR)));
            assertNotNull("key2 with faulty execution is missing.",
                    absoluteValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key2, ExecutionStatus.FAULT)));

        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getConvertedAbsoluteResponseTimeValues' of an
     * initialized probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>during the first sample: no absolute response time values is
     * available because no execution has been added,</li>
     * <li>after the switch of samples, no absolute response time values is
     * available because no execution has been added.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetConvertedAbsoluteResponseTimeValues_001() throws InterruptedException,
            ProbeException {
        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

        try {
            responseTimeProbe.init();

            assertEquals(
                    "No absolute values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getConvertedAbsoluteResponseTimeValues().size());

            // Wait more than one sample switch
            Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2.1));

            assertEquals(
                    "No absolute values should be available because no execution has been added.",
                    0, responseTimeProbe.getConvertedAbsoluteResponseTimeValues().size());

        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getConvertedAbsoluteResponseTimeValues' of a started
     * probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>during the first sample:
     * <ul></li>each response time is correctly collected</li>
     * <li>no absolute response time values is available</li>
     * <li>no relative response time values is available</li>
     * </ul>
     * </li>
     * <li>after the switch of samples, expected absolute response time values
     * are available as simple java type.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetConvertedAbsoluteResponseTimeValues_002() throws InterruptedException,
            ProbeException {
        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");
        final ProbeTestKey key2 = new ProbeTestKey("key2");

        try {
            responseTimeProbe.init();
            responseTimeProbe.start();

            final KeyedStartDateItem<ProbeTestKey> date1 = responseTimeProbe.newExecution(key1);
            responseTimeProbe.newExecution(key1);
            final KeyedStartDateItem<ProbeTestKey> date3 = responseTimeProbe.newExecution(key2);
            final KeyedStartDateItem<ProbeTestKey> date4 = responseTimeProbe.newExecution(key2);
            final KeyedStartDateItem<ProbeTestKey> date5 = responseTimeProbe.newExecution(key1);

            responseTimeProbe.endsExecution(date1, ExecutionStatus.SUCCEEDED);
            responseTimeProbe.endsExecution(date3, ExecutionStatus.ERROR);
            responseTimeProbe.endsExecution(date4, ExecutionStatus.FAULT);

            assertEquals(
                    "No absolute values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getConvertedAbsoluteResponseTimeValues().size());
            assertEquals(
                    "No relative values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getRelativeResponseTimeValues().size());

            // Wait more than one sample switch
            Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2));
            responseTimeProbe.endsExecution(date5, ExecutionStatus.SUCCEEDED);
            Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2));

            final Map<String[], Long[]> absoluteValues = responseTimeProbe
                    .getConvertedAbsoluteResponseTimeValues();
            assertEquals("Absolute values should be available after sample switch.", 4,
                    absoluteValues.size());

            Long[] value = null;
            for (final Map.Entry<String[], Long[]> entry : absoluteValues.entrySet()) {
                final String[] key = entry.getKey();
                if (key[0].equals("key1") && key[1].equals(ExecutionStatus.SUCCEEDED.toString())) {
                    value = entry.getValue();
                    break;
                }
            }
            assertTrue("key1 with successfully execution is missing.", value != null);
            assertTrue(
                    "Max response time of key1 with successfully execution is not the expected one.",
                    value[0] >= KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2);
            assertTrue(
                    "Min response time of key1 with successfully execution is not the expected one.",
                    value[2] <= KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

            boolean keyFound = false;
            for (final String[] key : absoluteValues.keySet()) {
                if (key[0].equals("key1") && key[1].equals(ExecutionStatus.PENDING.toString())) {
                    keyFound = true;
                }
            }
            assertTrue("key1 with pending execution is missing.", keyFound);

            keyFound = false;
            for (final String[] key : absoluteValues.keySet()) {
                if (key[0].equals("key2") && key[1].equals(ExecutionStatus.ERROR.toString())) {
                    keyFound = true;
                    break;
                }
            }
            assertTrue("key2 with execution in error is missing.", keyFound);

            keyFound = false;
            for (final String[] key : absoluteValues.keySet()) {
                if (key[0].equals("key2") && key[1].equals(ExecutionStatus.FAULT.toString())) {
                    keyFound = true;
                    break;
                }
            }
            assertTrue("key2 with faulty execution is missing.", keyFound);

        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getConvertedAbsoluteResponseTimeValues' of a stopped
     * probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>during the first sample:
     * <ul></li>each response time is correctly collected</li>
     * <li>no absolute response time values is available</li>
     * <li>no relative response time values is available</li>
     * </ul>
     * </li>
     * <li>after the switch of samples, expected absolute response time values
     * are available as simple java type.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetConvertedAbsoluteResponseTimeValues_003() throws InterruptedException,
            ProbeException {
        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");
        final ProbeTestKey key2 = new ProbeTestKey("key2");

        try {
            responseTimeProbe.init();
            responseTimeProbe.start();

            final KeyedStartDateItem<ProbeTestKey> date1 = responseTimeProbe.newExecution(key1);
            responseTimeProbe.newExecution(key1);
            final KeyedStartDateItem<ProbeTestKey> date3 = responseTimeProbe.newExecution(key2);
            final KeyedStartDateItem<ProbeTestKey> date4 = responseTimeProbe.newExecution(key2);
            final KeyedStartDateItem<ProbeTestKey> date5 = responseTimeProbe.newExecution(key1);

            responseTimeProbe.endsExecution(date1, ExecutionStatus.SUCCEEDED);
            responseTimeProbe.endsExecution(date3, ExecutionStatus.ERROR);
            responseTimeProbe.endsExecution(date4, ExecutionStatus.FAULT);

            assertEquals(
                    "No absolute values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getConvertedAbsoluteResponseTimeValues().size());
            assertEquals(
                    "No relative values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getRelativeResponseTimeValues().size());

            // Wait more than one sample switch
            Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2));
            responseTimeProbe.endsExecution(date5, ExecutionStatus.SUCCEEDED);
            responseTimeProbe.stop();
            Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2));

            final Map<String[], Long[]> absoluteValues = responseTimeProbe
                    .getConvertedAbsoluteResponseTimeValues();
            assertEquals("Absolute values should be available after sample switch.", 4,
                    absoluteValues.size());

            Long[] value = null;
            for (final Map.Entry<String[], Long[]> entry : absoluteValues.entrySet()) {
                final String[] key = entry.getKey();
                if (key[0].equals("key1") && key[1].equals(ExecutionStatus.SUCCEEDED.toString())) {
                    value = entry.getValue();
                    break;
                }
            }
            assertTrue("key1 with successfully execution is missing.", value != null);
            assertTrue(
                    "Max response time of key1 with successfully execution is not the expected one.",
                    value[0] >= KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2);
            assertTrue(
                    "Min response time of key1 with successfully execution is not the expected one.",
                    value[2] <= KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

            boolean keyFound = false;
            for (final String[] key : absoluteValues.keySet()) {
                if (key[0].equals("key1") && key[1].equals(ExecutionStatus.PENDING.toString())) {
                    keyFound = true;
                    break;
                }
            }
            assertTrue("key1 with pending execution is missing.", keyFound);

            keyFound = false;
            for (final String[] key : absoluteValues.keySet()) {
                if (key[0].equals("key2") && key[1].equals(ExecutionStatus.ERROR.toString())) {
                    keyFound = true;
                    break;
                }
            }
            assertTrue("key2 with execution in error is missing.", keyFound);

            keyFound = false;
            for (final String[] key : absoluteValues.keySet()) {
                if (key[0].equals("key2") && key[1].equals(ExecutionStatus.FAULT.toString())) {
                    keyFound = true;
                    break;
                }
            }
            assertTrue("key2 with faulty execution is missing.", keyFound);

        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getRelativeResponseTimeValues' of a started probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no exception is thrown,</li>
     * <li>during the first sample:
     * <ul></li>each response time is correctly collected</li>
     * <li>no absolute response time values is available</li>
     * <li>no relative response time values is available</li>
     * </ul>
     * </li>
     * <li>after the switch of samples, expected relative response time values
     * are available.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testGetRelativeResponseTimeValues_001() throws InterruptedException, ProbeException {
        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");
        final ProbeTestKey key2 = new ProbeTestKey("key2");

        try {
            responseTimeProbe.init();
            responseTimeProbe.start();

            final KeyedStartDateItem<ProbeTestKey> date1 = responseTimeProbe.newExecution(key1);
            responseTimeProbe.newExecution(key1);
            final KeyedStartDateItem<ProbeTestKey> date3 = responseTimeProbe.newExecution(key2);
            final KeyedStartDateItem<ProbeTestKey> date4 = responseTimeProbe.newExecution(key2);
            final KeyedStartDateItem<ProbeTestKey> date5 = responseTimeProbe.newExecution(key1);

            responseTimeProbe.endsExecution(date1, ExecutionStatus.SUCCEEDED);
            responseTimeProbe.endsExecution(date3, ExecutionStatus.ERROR);
            responseTimeProbe.endsExecution(date4, ExecutionStatus.FAULT);

            assertEquals(
                    "No absolute values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getAbsoluteResponseTimeValues().size());
            assertEquals(
                    "No relative values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getRelativeResponseTimeValues().size());

            // Wait that end of the current sample
            Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 1.1));
            Map<ExecutionStatusProbeKey<ProbeTestKey>, ResponseTimeRelativeValues> relativeValues = responseTimeProbe
                    .getRelativeResponseTimeValues();
            assertNotNull("key2 with execution in error is missing.",
                    relativeValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key2, ExecutionStatus.ERROR)));
            assertNotNull("key2 with faulty execution is missing.",
                    relativeValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key2, ExecutionStatus.FAULT)));

            responseTimeProbe.endsExecution(date5, ExecutionStatus.SUCCEEDED);
            // Wait that end of the current sample
            Thread.sleep(KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

            relativeValues = responseTimeProbe.getRelativeResponseTimeValues();
            assertEquals("Relative values should be available after sample switch.", 2,
                    relativeValues.size());
            final ResponseTimeRelativeValues responseTimeRelativeValues = relativeValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key1, ExecutionStatus.SUCCEEDED));
            assertNotNull("key1 with successfully execution is missing.",
                    responseTimeRelativeValues);
            assertTrue(
                    "Max response time of key1 with successfully execution is not the expected one.",
                    responseTimeRelativeValues.getMax() >= KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

            assertTrue(
                    "Min response time of key1 with successfully execution is not the expected one.",
                    responseTimeRelativeValues.getMin() == responseTimeRelativeValues.getMax());
            assertNotNull("key1 with pending execution is missing.",
                    relativeValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key1, ExecutionStatus.PENDING)));

        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests that relative response time values are merged with the absolute
     * response time values when stopping the probe.
     * </p>
     * <p>
     * Expected results: the relative response time values are merged with the
     * absolute response time values
     * </p>
     */
    @Test
    public void testGetRelativeResponseTimeValues_002() throws InterruptedException, ProbeException {
        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");
        final ProbeTestKey key2 = new ProbeTestKey("key2");

        try {
            responseTimeProbe.init();
            responseTimeProbe.start();

            final KeyedStartDateItem<ProbeTestKey> date1 = responseTimeProbe.newExecution(key1);
            responseTimeProbe.newExecution(key1);
            final KeyedStartDateItem<ProbeTestKey> date3 = responseTimeProbe.newExecution(key2);
            final KeyedStartDateItem<ProbeTestKey> date4 = responseTimeProbe.newExecution(key2);
            final KeyedStartDateItem<ProbeTestKey> date5 = responseTimeProbe.newExecution(key1);

            responseTimeProbe.endsExecution(date1, ExecutionStatus.SUCCEEDED);
            responseTimeProbe.endsExecution(date3, ExecutionStatus.ERROR);
            responseTimeProbe.endsExecution(date4, ExecutionStatus.FAULT);

            assertEquals(
                    "No absolute values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getAbsoluteResponseTimeValues().size());
            assertEquals(
                    "No relative values should be available because no sample switch has had time to occur.",
                    0, responseTimeProbe.getRelativeResponseTimeValues().size());

            // Wait more than one sample switch
            Thread.sleep(KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2);
            responseTimeProbe.endsExecution(date5, ExecutionStatus.SUCCEEDED);
            responseTimeProbe.stop();
            Thread.sleep(KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2);

            final Map<ExecutionStatusProbeKey<ProbeTestKey>, AbsoluteResponseTimes> absoluteValues = responseTimeProbe
                    .getAbsoluteResponseTimeValues();
            assertEquals("Absolute values should be available after sample switch.", 4,
                    absoluteValues.size());
            assertNotNull("key1 with successfully execution is missing.",
                    absoluteValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key1, ExecutionStatus.SUCCEEDED)));
            assertTrue(
                    "Max response time of key1 with successfully execution is not the expected one.",
                    absoluteValues
                            .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key1,
                                    ExecutionStatus.SUCCEEDED)).getMaxResponseTime() >= KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2);

            assertTrue(
                    "Min response time of key1 with successfully execution is not the expected one.",
                    absoluteValues
                            .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key1,
                                    ExecutionStatus.SUCCEEDED)).getMinResponseTime() <= KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
            assertNotNull("key1 with pending execution is missing.",
                    absoluteValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key1, ExecutionStatus.PENDING)));
            assertNotNull("key2 with execution in error is missing.",
                    absoluteValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key2, ExecutionStatus.ERROR)));
            assertNotNull("key2 with faulty execution is missing.",
                    absoluteValues
                    .get(new ExecutionStatusProbeKeyImpl<ProbeKey>(key2, ExecutionStatus.FAULT)));

        } finally {
            timer.cancel();
        }
    }

    // TODO: Add unit tests about 'getConvertedAbsoluteResponseTimeValues'
    // TODO: Add unit tests about 'getConvertedRelativeResponseTimeValues'

    /*----------------------------*/
    /* Lifecycle error test cases */
    /*----------------------------*/

    /**
     * <p>
     * Tests the method 'newExecution' of a just allocated probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>no absolute response time value is available, even after sample
     * switch,</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testNewExecution_101() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");

        try {
            try {
                responseTimeProbe.newExecution(key1);
                fail("Invoking 'newExecution' of a just allocated probe should fail with ProbeNotStartedException");
            } catch (final ProbeNotStartedException e) {
                // NOP, it's the expected exception
                assertEquals(
                        "No absolute values should be available because no sample switch has had time to occur.",
                        0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

                // Wait the sample switch
                Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 1.1));

                assertEquals(
                        "No absolute values should be available after sample switch.", 0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

                throw e;
            }
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'newExecution' of an initialized probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>no absolute response time value is available, even after sample
     * switch,</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testNewExecution_102() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");

        try {
            try {
                responseTimeProbe.init();

                responseTimeProbe.newExecution(key1);
                fail("Invoking 'newExecution' of an initialized probe should fail with ProbeNotStartedException");
            } catch (final ProbeNotStartedException e) {
                // NOP, it's the expected exception
                assertEquals(
                        "No absolute values should be available because no sample switch has had time to occur.",
                        0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

                // Wait the sample switch
                Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 1.1));

                assertEquals("Absolute values should be available after sample switch.",
                        0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

                throw e;
            }
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'newExecution' of a stopped probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>absolute response time values are available after sample switch,</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testNewExecution_103() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");

        try {
            try {
                responseTimeProbe.init();
                responseTimeProbe.start();
                responseTimeProbe.newExecution(key1);
                assertEquals(
                        "No absolute values should be available because no sample switch has had time to occur.",
                        0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());
                // Wait the sample switches
                Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 1.1));
                assertEquals("Absolute values should be available after sample switch.",
                        1,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());
                responseTimeProbe.stop();

                responseTimeProbe.newExecution(key1);
                fail("Invoking 'newExecution' of a stopped probe should fail with ProbeNotStartedException");
            } catch (final ProbeNotStartedException e) {
                // NOP, it's the expected exception
                assertEquals(
                        "Absolute values should be available when the probe is stopped.", 1,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());
                throw e;
            }
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'newExecution' of an shutdown probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>no absolute response time values are available after sample switch,</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testNewExecution_104() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");

        try {
            try {
                responseTimeProbe.init();
                responseTimeProbe.start();
                responseTimeProbe.newExecution(key1);
                assertEquals(
                        "No absolute values should be available because no sample switch has had time to occur.",
                        0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());
                // Wait the sample switch
                Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 1.1));
                assertEquals("Absolute values should be available after sample switch.",
                        1,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());
                // Wait another sample switch
                responseTimeProbe.stop();
                responseTimeProbe.shutdown();

                responseTimeProbe.newExecution(key1);
                fail("Invoking 'newExecution' of a shutdown probe should fail with ProbeNotStartedException");
            } catch (final ProbeNotStartedException e) {
                // NOP, it's the expected exception
                assertEquals(
                        "No absolute values should be available when the probe is shutdown.", 0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

                throw e;
            }
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'endsExecution' of a just allocated probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>no absolute response time value is available, even after sample
     * switch,</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testEndsExecution_101() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

        try {
            try {
                responseTimeProbe.endsExecution(new KeyedStartDateItem<ProbeTestKey>(
                        new ProbeTestKey(
                        "key1"), System.currentTimeMillis()), ExecutionStatus.SUCCEEDED);
                fail("Invoking 'endsExecution' of a just allocated probe should fail with ProbeNotStartedException");
            } catch (final ProbeNotStartedException e) {
                // NOP, it's the expected exception
                assertEquals(
                        "No absolute values should be available because no sample switch has had time to occur.",
                        0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

                // Wait the sample switch
                Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 1.1));

                assertEquals(
                        "No absolute values should be available after sample switch.", 0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

                throw e;
            }
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'endsExecution' of an initialized probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>no absolute response time value is available, even after sample
     * switch,</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testEndsExecution_102() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

        try {
            try {
                responseTimeProbe.init();

                responseTimeProbe.endsExecution(new KeyedStartDateItem<ProbeTestKey>(
                        new ProbeTestKey(
                        "key1"), System.currentTimeMillis()), ExecutionStatus.SUCCEEDED);
                fail("Invoking 'endsExecution' of an initialized probe should fail with ProbeNotStartedException");
            } catch (final ProbeNotStartedException e) {
                // NOP, it's the expected exception
                assertEquals(
                        "No absolute values should be available because no sample switch has had time to occur.",
                        0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

                // Wait the sample switch
                Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 1.1));

                assertEquals("Absolute values should be available after sample switch.",
                        0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

                throw e;
            }
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'endsExecution' of a stopped probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>absolute response time values are available after sample switch,</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testEndsExecution_103() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");

        try {
            try {
                responseTimeProbe.init();
                responseTimeProbe.start();
                final KeyedStartDateItem<ProbeTestKey> startDate1 = responseTimeProbe
                        .newExecution(key1);
                responseTimeProbe.endsExecution(startDate1, ExecutionStatus.SUCCEEDED);
                assertEquals(
                        "No absolute values should be available because no sample switch has had time to occur.",
                        0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());
                // Wait the sample switches
                Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 1.1));
                assertEquals("Absolute values should be available after sample switch.",
                        1,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());
                responseTimeProbe.stop();

                responseTimeProbe.endsExecution(new KeyedStartDateItem<ProbeTestKey>(
                        new ProbeTestKey(
                        "key1"), System.currentTimeMillis()), ExecutionStatus.SUCCEEDED);
                fail("Invoking 'endsExecution' of a stopped probe should fail with ProbeNotStartedException");
            } catch (final ProbeNotStartedException e) {
                // NOP, it's the expected exception
                assertEquals(
                        "Absolute values should be available when the probe is stopped.", 1,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

                throw e;
            }
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'endsExecution' of an shutdown probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotStartedException} is thrown,</li>
     * <li>no absolute response time values are available after sample switch,</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testEndsExecution_104() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");

        try {
            try {
                responseTimeProbe.init();
                responseTimeProbe.start();
                final KeyedStartDateItem<ProbeTestKey> startDate1 = responseTimeProbe
                        .newExecution(key1);
                assertEquals(
                        "No absolute values should be available because no sample switch has had time to occur.",
                        0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());
                // Wait the sample switch
                Thread.sleep(KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2);
                assertEquals("Absolute values should be available after sample switch.",
                        1,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());
                // Wait another sample switch
                responseTimeProbe.stop();
                responseTimeProbe.shutdown();

                responseTimeProbe.endsExecution(startDate1, ExecutionStatus.SUCCEEDED);
                fail("Invoking 'endsExecution' of a shutdown probe should fail with ProbeNotStartedException");
            } catch (final ProbeNotStartedException e) {
                // NOP, it's the expected exception
                assertEquals(
                        "No absolute values should be available when the probe is shutdown.", 0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

                throw e;
            }
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'endsExecution' with a start date that was not
     * previously returned by 'newExecution'.
     * </p>
     * <p>
     * Expected results: the exception {@link StartDateItemUnknownException} is
     * thrown.
     * </p>
     */
    @Test(expected = StartDateItemUnknownException.class)
    public void testEndsExecution_105() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");

        try {
            responseTimeProbe.init();
            responseTimeProbe.start();
            responseTimeProbe.endsExecution(
                    new KeyedStartDateItem<ProbeTestKey>(key1, System.currentTimeMillis()),
                    ExecutionStatus.SUCCEEDED);
            fail("Invoking 'endsExecution' with a start date not returned by 'newExecution' should fail with StartDateItemUnknownException");
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getAbsoluteResponseTimeValues' of a just allocated
     * probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotInitializedException} is thrown,</li>
     * <li>no absolute response time value is available, even after sample
     * switch,</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotInitializedException.class)
    public void testGetAbsoluteResponseTimeValues_101() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

        try {
            responseTimeProbe.getAbsoluteResponseTimeValues();
            fail("Invoking 'getAbsoluteResponseTimeValues' of a just allocated probe should fail with ProbeNotInitializedException");
        } catch (final ProbeNotInitializedException e) {
            // NOP, it's the expected exception
            assertEquals(
                    "No absolute values should be available because no sample switch has had time to occur.",
                    0,
                    ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

            // Wait the sample switch
            Thread.sleep(KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2);

            assertEquals("No absolute values should be available after sample switch.", 0,
                    ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

            throw e;
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getAbsoluteResponseTimeValues' of an shutdown probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotInitializedException} is thrown,</li>
     * <li>no absolute response time values are available after sample switch,</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotInitializedException.class)
    public void testGetAbsoluteResponseTimeValues_102() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");

        try {
            try {
                responseTimeProbe.init();
                responseTimeProbe.start();
                responseTimeProbe.newExecution(key1);
                assertEquals(
                        "No absolute values should be available because no sample switch has had time to occur.",
                        0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());
                // Wait the sample switch
                Thread.sleep(KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2);
                assertEquals("Absolute values should be available after sample switch.",
                        1,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());
                // Wait another sample switch
                responseTimeProbe.stop();
                responseTimeProbe.shutdown();

                responseTimeProbe.getAbsoluteResponseTimeValues();
                fail("Invoking 'getAbsoluteResponseTimeValues' of a shutdown probe should fail with ProbeNotInitializedException");
            } catch (final ProbeNotInitializedException e) {
                // NOP, it's the expected exception
                assertEquals(
                        "No absolute values should be available when the probe is shutdown.", 0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

                throw e;
            }
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getConvertedAbsoluteResponseTimeValues' of a just
     * allocated probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotInitializedException} is thrown,</li>
     * <li>no absolute response time value is available, even after sample
     * switch,</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotInitializedException.class)
    public void testGetConvertedAbsoluteResponseTimeValues_101() throws InterruptedException,
            ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

        try {
            responseTimeProbe.getConvertedAbsoluteResponseTimeValues();
            fail("Invoking 'getConvertedAbsoluteResponseTimeValues' of a just allocated probe should fail with ProbeNotInitializedException");
        } catch (final ProbeNotInitializedException e) {
            // NOP, it's the expected exception
            assertEquals(
                    "No absolute values should be available because no sample switch has had time to occur.",
                    0,
                    ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

            // Wait the sample switch
            Thread.sleep(KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2);

            assertEquals("No absolute values should be available after sample switch.", 0,
                    ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

            throw e;
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getConvertedAbsoluteResponseTimeValues' of an shutdown
     * probe.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>the exception {@link ProbeNotInitializedException} is thrown,</li>
     * <li>no absolute response time values are available after sample switch,</li>
     * </ul>
     * </p>
     */
    @Test(expected = ProbeNotInitializedException.class)
    public void testGetConvertedAbsoluteResponseTimeValues_102() throws InterruptedException,
            ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        final ProbeTestKey key1 = new ProbeTestKey("key1");

        try {
            try {
                responseTimeProbe.init();
                responseTimeProbe.start();
                responseTimeProbe.newExecution(key1);
                assertEquals(
                        "No absolute values should be available because no sample switch has had time to occur.",
                        0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());
                // Wait the sample switch
                Thread.sleep(KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 2);
                assertEquals("Absolute values should be available after sample switch.",
                        1,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());
                // Wait another sample switch
                responseTimeProbe.stop();
                responseTimeProbe.shutdown();

                responseTimeProbe.getConvertedAbsoluteResponseTimeValues();
                fail("Invoking 'getConvertedAbsoluteResponseTimeValues' of a shutdown probe should fail with ProbeNotInitializedException");
            } catch (final ProbeNotInitializedException e) {
                // NOP, it's the expected exception
                assertEquals(
                        "No absolute values should be available when the probe is shutdown.", 0,
                        ((Map<?, AbsoluteResponseTimes>) ReflectionHelper.getPrivateFieldValue(KeyedResponseTimeProbeImpl.class, responseTimeProbe, "absoluteResponseTimeValues")).size());

                throw e;
            }
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getRelativeResponseTimeValues' of a just allocated
     * probe.
     * </p>
     * <p>
     * Expected results: the exception {@link ProbeNotStartedException} is
     * thrown.
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testGetRelativeResponseTimeValues_101() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

        try {
            responseTimeProbe.getRelativeResponseTimeValues();
            fail("Invoking 'getRelativeResponseTimeValues' of a just allocated probe should fail with ProbeNotStartedException");
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getRelativeResponseTimeValues' of an initialized probe.
     * </p>
     * <p>
     * Expected results: the exception {@link ProbeNotStartedException} is
     * thrown.
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testGetRelativeResponseTimeValues_102() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

        try {
            responseTimeProbe.init();

            responseTimeProbe.getRelativeResponseTimeValues();
            fail("Invoking 'getRelativeResponseTimeValues' of an initialized probe should fail with ProbeNotStartedException");
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getRelativeResponseTimeValues' of a stopped probe.
     * </p>
     * <p>
     * Expected results: the exception {@link ProbeNotStartedException} is
     * thrown.
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testGetRelativeResponseTimeValues_103() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

        try {
            responseTimeProbe.init();
            responseTimeProbe.stop();

            responseTimeProbe.getRelativeResponseTimeValues();
            fail("Invoking 'getRelativeResponseTimeValues' of a stopped probe should fail with ProbeNotStartedException");
        } finally {
            timer.cancel();
        }
    }

    /**
     * <p>
     * Tests the method 'getRelativeResponseTimeValues' of a shutdown probe.
     * </p>
     * <p>
     * Expected results: the exception {@link ProbeNotStartedException} is
     * thrown.
     * </p>
     */
    @Test(expected = ProbeNotStartedException.class)
    public void testGetRelativeResponseTimeValues_104() throws InterruptedException, ProbeException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

        try {
            responseTimeProbe.init();
            responseTimeProbe.stop();
            responseTimeProbe.shutdown();

            responseTimeProbe.getRelativeResponseTimeValues();
            fail("Invoking 'getRelativeResponseTimeValues' of a shutdown probe should fail with ProbeNotStartedException");
        } finally {
            timer.cancel();
        }
    }

    /*--------------------------*/
    /* Thread-safety test cases */
    /*--------------------------*/

    /**
     * <p>
     * Tests the thread-safety of the methods. No change is made on the probe
     * lifecycle.
     * </p>
     */
    @Test
    public void testThreadSafety_001() throws ProbeException, InterruptedException {

        // Testcase failures
        final List<Throwable> testcaseFailures = Collections
                .synchronizedList(new LinkedList<Throwable>());

        // Synchronization objects
        final Object starterMonitor = new Object();
        final Semaphore readyMonitor = new Semaphore(THREAD_POOL_SIZE);

        // Initialize and start the probe
        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);

        try {
            responseTimeProbe.init();
            responseTimeProbe.start();

            // Initialize and start the thread pool size to THREAD_POOL_SIZE
            final List<Thread> threads = new LinkedList<Thread>();
            for (int k = 0; k < KEYS_NUMBER; k++) {
                final ProbeTestKey key = new ProbeTestKey("key" + k);
                for (int i = 0; i < THREAD_POOL_SIZE; i++) {
                    readyMonitor.acquire();

                    final Thread thread = new Thread(new Runnable() {

                        @Override
                        public void run() {

                            try {
                                synchronized (starterMonitor) {
                                    readyMonitor.release();
                                    starterMonitor.wait();
                                }

                                for (int j = 0; j < THREADSAFETY_ITERATION_NB; j++) {
                                    final KeyedStartDateItem<ProbeTestKey> startDate = responseTimeProbe
                                            .newExecution(key);
                                    responseTimeProbe.endsExecution(startDate,
                                            ExecutionStatus.SUCCEEDED);

                                    responseTimeProbe.getAbsoluteResponseTimeValues();
                                    responseTimeProbe.getRelativeResponseTimeValues();
                                }
                            } catch (final Throwable e) {
                                testcaseFailures.add(e);
                            }

                        }
                    });
                    threads.add(thread);
                    thread.start();
                }
            }

            // Wait that threads end to start
            readyMonitor.acquire(THREAD_POOL_SIZE);

            synchronized (starterMonitor) {
                starterMonitor.notifyAll();
            }

            // Wait the end of processing
            for (Thread thread : threads) {
                thread.join();
            }

            // Wait the last sample switch
            Thread.sleep((long) (KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD * 1.1));

            assertEquals("Error occurs: " + testcaseFailures.toString(), 0,
                    testcaseFailures.size());

            for (int k = 0; k < KEYS_NUMBER; k++) {
                final AbsoluteResponseTimes absoluteValue = responseTimeProbe
                        .getAbsoluteResponseTimeValues().get(
                                new ExecutionStatusProbeKeyImpl<ProbeKey>(new ProbeTestKey("key" + k),
                                        ExecutionStatus.SUCCEEDED));
                assertEquals("The expected response time number is not the expected one",
                        THREAD_POOL_SIZE * THREADSAFETY_ITERATION_NB,
                        ((Long)ReflectionHelper.getPrivateFieldValue(AbsoluteResponseTimes.class, absoluteValue, "responseTimesNb")).longValue());
            }
        } finally {
            timer.cancel();
        }
    }

    /**
     * Check that no error occurs on start/stop cycles (see PETALSCDK-116)
     */
    @Test
    public void startStopCycles() throws ProbeInitializedException, ProbeStartedException,
            ProbeInitializationException, ProbeNotInitializedException, ProbeNotStartedException, ProbeStopException,
            ProbeStartupException, ProbeShutdownException {

        final Timer timer = new Timer();
        final KeyedResponseTimeProbe<ProbeTestKey> responseTimeProbe = new KeyedResponseTimeProbeImpl<ProbeTestKey>(
                timer, KeyedResponseTimeProbeImplTestCase.SAMPLE_PERIOD);
        try {
            responseTimeProbe.init();

            responseTimeProbe.start();
            responseTimeProbe.stop();

            responseTimeProbe.start();
            responseTimeProbe.stop();

            responseTimeProbe.shutdown();
        } finally {
            timer.cancel();
        }
    }

    // TODO: Add tests about thread-safety of methods and stop collecting

    // TODO: Add tests about thread-safety of methods and lifecycle changes
}
