/**
 * Copyright (c) 2013-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.Semaphore;

import junit.framework.TestCase;

import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.ProbeException;
import org.ow2.petals.probes.api.probes.ResponseTimeRelativeValues;
import org.ow2.petals.probes.api.probes.StartDateItem;

/**
 * <p>
 * Unit tests of {@link ResponseTimesSample} to verify its thread-safety.
 * </p>
 * 
 * @author Christophe DENEUX (PetalsLink)
 * 
 */
public class ResponseTimesSampleTestCase extends TestCase {

    private final static short THREAD_POOL_SIZE = 25;

    private final static long THREADSAFETY_ITERATION_NB = 4000;

    /*--------------------------*/
    /* Thread-safety test cases */
    /*--------------------------*/

    /**
     * <p>
     * Tests the thread-safety of the methods 'inc' and 'getValues'. No change
     * is made on the probe lifecycle.
     * </p>
     * 
     * @throws InterruptedException
     */
    public void testThreadSafety_001() throws ProbeException, InterruptedException {

        // Testcase failures
        final List<Throwable> testcaseFailures = Collections
                .synchronizedList(new LinkedList<Throwable>());

        // Synchronization objects
        final Object starterMonitor = new Object();
        final Semaphore readyMonitor = new Semaphore(THREAD_POOL_SIZE);

        final ResponseTimesSample responseTimesSample = new ResponseTimesSample();
        final BlockingQueue<StartDateItem> endsExecutionQueue = new LinkedBlockingQueue<StartDateItem>();

        final List<Thread> threads = new LinkedList<Thread>();
        for (int i = 0; i < THREAD_POOL_SIZE; i++) {
            readyMonitor.acquire();
            // Threads that start execution
            final Thread addStartThread = new Thread(new Runnable() {

                @Override
                public void run() {

                    try {
                        synchronized (starterMonitor) {
                            readyMonitor.release();
                            starterMonitor.wait();
                        }

                        for (int j = 0; j < THREADSAFETY_ITERATION_NB; j++) {
                            final StartDateItem startDate = responseTimesSample.addStartDate(System
                                    .currentTimeMillis());
                            endsExecutionQueue.add(startDate);
                        }

                    } catch (final Throwable e) {
                        testcaseFailures.add(e);
                    }
                }
            });
            threads.add(addStartThread);
            addStartThread.start();

            // Threads that call 'move'.
            final Thread endsExecutionThread = new Thread(new Runnable() {

                @Override
                public void run() {

                    try {
                        synchronized (starterMonitor) {
                            readyMonitor.release();
                            starterMonitor.wait();
                        }

                        int j = 0;
                        while (j < THREADSAFETY_ITERATION_NB) {
                            final StartDateItem startDate = endsExecutionQueue.poll();
                            if (startDate != null) {
                                Thread.sleep((long) (Math.random() * 50));
                                responseTimesSample.endStartDate(startDate,
                                        System.currentTimeMillis(), ExecutionStatus.SUCCEEDED);
                                j++;
                            }
                        }

                    } catch (final Throwable e) {
                        testcaseFailures.add(e);
                    }
                }
            });
            threads.add(endsExecutionThread);
            endsExecutionThread.start();
        }

        // Wait that threads end to start
        readyMonitor.acquire(THREAD_POOL_SIZE * 2);

        synchronized (starterMonitor) {
            starterMonitor.notifyAll();
        }

        // Wait the end of processing
        for (Thread thread : threads) {
            thread.join();
        }

        TestCase.assertEquals("Error occurs: " + testcaseFailures.toString(), 0,
                testcaseFailures.size());

        responseTimesSample.stopCollecting();

        final Map<ExecutionStatus, ResponseTimeRelativeValues> responseTimeValuesList = responseTimesSample
                .getResponseTimeValues();
        final ResponseTimeRelativeValues responseTimeValues = responseTimeValuesList
                .get(ExecutionStatus.SUCCEEDED);
        TestCase.assertNotNull("Missing response time values for executin status '"
                + ExecutionStatus.SUCCEEDED.name() + "'.", responseTimeValues);
        TestCase.assertEquals("Unexpected number of response times for executin status '"
                + ExecutionStatus.SUCCEEDED.name() + "'.", THREAD_POOL_SIZE
                * THREADSAFETY_ITERATION_NB, responseTimeValues.getResponseTimesNb());
    }
}
